# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

import lightning.pytorch as pl
import torch
from megatron.core.optimizer import OptimizerConfig

from nemo import lightning as nl
from nemo.collections import llm
from nemo.lightning.io.mixin import track_io


def get_args():
    parser = argparse.ArgumentParser(description='Finetune a small GPT model using NeMo 2.0')
    parser.add_argument('--model', type=str.lower, choices=['mistral', 'mixtral'], help="model")
    parser.add_argument('--max-steps', type=int, default=9, help="number of devices")
    parser.add_argument('--mbs', type=int, default=2, help="micro batch size")
    parser.add_argument('--gbs', type=int, default=4, help="global batch size")
    parser.add_argument('--tp', type=int, default=1, help="tensor parallel size")
    parser.add_argument('--ep', type=int, default=1, help="expert parallel size")
    parser.add_argument('--dist-opt', action='store_true', help='use dist opt')
    parser.add_argument('--use-exclude', action='store_true', help='will use exclude_modules')
    return parser.parse_args()


def trainer(devices, tp, ep, sp, max_steps) -> nl.Trainer:
    strategy = nl.MegatronStrategy(
        tensor_model_parallel_size=tp,
        expert_model_parallel_size=ep,
        sequence_parallel=sp,
    )

    return nl.Trainer(
        devices=max(ep, tp),
        max_steps=max_steps,
        accelerator="gpu",
        strategy=strategy,
        plugins=nl.MegatronMixedPrecision(precision="bf16-mixed"),
        log_every_n_steps=1,
        limit_val_batches=0,
        val_check_interval=0,
        num_sanity_val_steps=0,
    )


@track_io
class OrdTokenizer:
    def __init__(self, vocab_size=30_000, num_reserved_tokens=128, special_token_names=['bos_id', 'eos_id', 'pad_id']):
        self.vocab_size = vocab_size
        self.num_reserved_tokens = num_reserved_tokens
        self.special_token_names = special_token_names
        assert len(self.special_token_names) < num_reserved_tokens

    def __getattr__(self, name):
        if name in self.__dict__.get('special_token_names', {}):
            return self.__dict__['special_token_names'].index(name)
        elif name in self.__dict__:
            return self.__dict__[name]
        else:
            raise AttributeError

    def text_to_ids(self, text):
        token_ids = list(map(lambda x: self.num_reserved_tokens + ord(x), list(text)))
        assert max(token_ids) < self.vocab_size
        return token_ids


def logger() -> nl.NeMoLogger:
    ckpt = nl.ModelCheckpoint(
        save_last=True,
        every_n_train_steps=10,
        monitor="reduced_train_loss",
        save_top_k=1,
        save_on_train_epoch_end=True,
        save_optim_on_train_end=True,
    )

    return nl.NeMoLogger(
        name="nemo2_peft",
        log_dir="/tmp/peft_logs",
        use_datetime_version=False,  # must be false if using auto resume
        ckpt=ckpt,
        wandb=None,
    )


def squad(mbs, gbs) -> pl.LightningDataModule:
    return llm.SquadDataModule(seq_length=2048, micro_batch_size=mbs, global_batch_size=gbs, num_workers=0)


def get_mistral_expected_ckpt():
    return {
        "module.embedding.word_embeddings.weight": ("Tensor", [15104, 4096]),
        "module.decoder.layers.0.self_attention.core_attention._extra_state": ("Tensor", [4]),
        "module.decoder.layers.0.self_attention.linear_proj.weight": ("Tensor", [4096, 2048]),
        "module.decoder.layers.0.self_attention.linear_proj._extra_state": ("Tensor", [4]),
        "module.decoder.layers.0.self_attention.linear_proj.adapter.linear_in.weight": ("Tensor", [32, 2048]),
        "module.decoder.layers.0.self_attention.linear_proj.adapter.linear_in._extra_state": ("NoneType", []),
        "module.decoder.layers.0.self_attention.linear_proj.adapter.linear_out.weight": ("Tensor", [2048, 32]),
        "module.decoder.layers.0.self_attention.linear_proj.adapter.linear_out._extra_state": ("NoneType", []),
        "module.decoder.layers.0.self_attention.linear_qkv.layer_norm_weight": ("Tensor", [4096]),
        "module.decoder.layers.0.self_attention.linear_qkv.weight": ("Tensor", [3072, 4096]),
        "module.decoder.layers.0.self_attention.linear_qkv._extra_state": ("Tensor", [4]),
        "module.decoder.layers.0.self_attention.linear_qkv.adapter.linear_in.weight": ("Tensor", [16, 4096]),
        "module.decoder.layers.0.self_attention.linear_qkv.adapter.linear_in._extra_state": ("NoneType", []),
        "module.decoder.layers.0.self_attention.linear_qkv.adapter.linear_out.weight": ("Tensor", [3072, 32]),
        "module.decoder.layers.0.self_attention.linear_qkv.adapter.linear_out._extra_state": ("NoneType", []),
        "module.decoder.layers.0.mlp.linear_fc1.layer_norm_weight": ("Tensor", [4096]),
        "module.decoder.layers.0.mlp.linear_fc1.weight": ("Tensor", [14336, 4096]),
        "module.decoder.layers.0.mlp.linear_fc1._extra_state": ("Tensor", [4]),
        "module.decoder.layers.0.mlp.linear_fc1.adapter.linear_in.weight": ("Tensor", [16, 4096]),
        "module.decoder.layers.0.mlp.linear_fc1.adapter.linear_in._extra_state": ("NoneType", []),
        "module.decoder.layers.0.mlp.linear_fc1.adapter.linear_out.weight": ("Tensor", [14336, 32]),
        "module.decoder.layers.0.mlp.linear_fc1.adapter.linear_out._extra_state": ("NoneType", []),
        "module.decoder.layers.0.mlp.linear_fc2.weight": ("Tensor", [4096, 7168]),
        "module.decoder.layers.0.mlp.linear_fc2._extra_state": ("Tensor", [4]),
        "module.decoder.layers.0.mlp.linear_fc2.adapter.linear_in.weight": ("Tensor", [32, 7168]),
        "module.decoder.layers.0.mlp.linear_fc2.adapter.linear_in._extra_state": ("NoneType", []),
        "module.decoder.layers.0.mlp.linear_fc2.adapter.linear_out.weight": ("Tensor", [2048, 32]),
        "module.decoder.layers.0.mlp.linear_fc2.adapter.linear_out._extra_state": ("NoneType", []),
        "module.decoder.layers.1.self_attention.core_attention._extra_state": ("Tensor", [4]),
        "module.decoder.layers.1.self_attention.linear_proj.weight": ("Tensor", [4096, 2048]),
        "module.decoder.layers.1.self_attention.linear_proj._extra_state": ("Tensor", [4]),
        "module.decoder.layers.1.self_attention.linear_proj.adapter.linear_in.weight": ("Tensor", [32, 2048]),
        "module.decoder.layers.1.self_attention.linear_proj.adapter.linear_in._extra_state": ("NoneType", []),
        "module.decoder.layers.1.self_attention.linear_proj.adapter.linear_out.weight": ("Tensor", [2048, 32]),
        "module.decoder.layers.1.self_attention.linear_proj.adapter.linear_out._extra_state": ("NoneType", []),
        "module.decoder.layers.1.self_attention.linear_qkv.layer_norm_weight": ("Tensor", [4096]),
        "module.decoder.layers.1.self_attention.linear_qkv.weight": ("Tensor", [3072, 4096]),
        "module.decoder.layers.1.self_attention.linear_qkv._extra_state": ("Tensor", [4]),
        "module.decoder.layers.1.self_attention.linear_qkv.adapter.linear_in.weight": ("Tensor", [16, 4096]),
        "module.decoder.layers.1.self_attention.linear_qkv.adapter.linear_in._extra_state": ("NoneType", []),
        "module.decoder.layers.1.self_attention.linear_qkv.adapter.linear_out.weight": ("Tensor", [3072, 32]),
        "module.decoder.layers.1.self_attention.linear_qkv.adapter.linear_out._extra_state": ("NoneType", []),
        "module.decoder.layers.1.mlp.linear_fc1.layer_norm_weight": ("Tensor", [4096]),
        "module.decoder.layers.1.mlp.linear_fc1.weight": ("Tensor", [14336, 4096]),
        "module.decoder.layers.1.mlp.linear_fc1._extra_state": ("Tensor", [4]),
        "module.decoder.layers.1.mlp.linear_fc1.adapter.linear_in.weight": ("Tensor", [16, 4096]),
        "module.decoder.layers.1.mlp.linear_fc1.adapter.linear_in._extra_state": ("NoneType", []),
        "module.decoder.layers.1.mlp.linear_fc1.adapter.linear_out.weight": ("Tensor", [14336, 32]),
        "module.decoder.layers.1.mlp.linear_fc1.adapter.linear_out._extra_state": ("NoneType", []),
        "module.decoder.layers.1.mlp.linear_fc2.weight": ("Tensor", [4096, 7168]),
        "module.decoder.layers.1.mlp.linear_fc2._extra_state": ("Tensor", [4]),
        "module.decoder.layers.1.mlp.linear_fc2.adapter.linear_in.weight": ("Tensor", [32, 7168]),
        "module.decoder.layers.1.mlp.linear_fc2.adapter.linear_in._extra_state": ("NoneType", []),
        "module.decoder.layers.1.mlp.linear_fc2.adapter.linear_out.weight": ("Tensor", [2048, 32]),
        "module.decoder.layers.1.mlp.linear_fc2.adapter.linear_out._extra_state": ("NoneType", []),
        "module.decoder.final_layernorm.weight": ("Tensor", [4096]),
        "module.decoder.final_layernorm._extra_state": ("Tensor", [0]),
        "module.output_layer.weight": ("Tensor", [15104, 4096]),
        "module.output_layer._extra_state": ("NoneType", []),
    }


def get_mixtral_expected_ckpt():
    return {
        "module.embedding.word_embeddings.weight": ("Tensor", [15104, 4096]),
        "module.decoder.layers.0.self_attention.core_attention._extra_state": ("Tensor", [4]),
        "module.decoder.layers.0.self_attention.linear_proj.weight": ("Tensor", [4096, 2048]),
        "module.decoder.layers.0.self_attention.linear_proj._extra_state": ("Tensor", [4]),
        "module.decoder.layers.0.self_attention.linear_proj.adapter.linear_in.weight": ("Tensor", [32, 2048]),
        "module.decoder.layers.0.self_attention.linear_proj.adapter.linear_in._extra_state": ("NoneType", []),
        "module.decoder.layers.0.self_attention.linear_proj.adapter.linear_out.weight": ("Tensor", [2048, 32]),
        "module.decoder.layers.0.self_attention.linear_proj.adapter.linear_out._extra_state": ("NoneType", []),
        "module.decoder.layers.0.self_attention.linear_qkv.layer_norm_weight": ("Tensor", [4096]),
        "module.decoder.layers.0.self_attention.linear_qkv.weight": ("Tensor", [3072, 4096]),
        "module.decoder.layers.0.self_attention.linear_qkv._extra_state": ("Tensor", [4]),
        "module.decoder.layers.0.self_attention.linear_qkv.adapter.linear_in.weight": ("Tensor", [16, 4096]),
        "module.decoder.layers.0.self_attention.linear_qkv.adapter.linear_in._extra_state": ("NoneType", []),
        "module.decoder.layers.0.self_attention.linear_qkv.adapter.linear_out.weight": ("Tensor", [3072, 32]),
        "module.decoder.layers.0.self_attention.linear_qkv.adapter.linear_out._extra_state": ("NoneType", []),
        "module.decoder.layers.0.pre_mlp_layernorm.weight": ("Tensor", [4096]),
        "module.decoder.layers.0.pre_mlp_layernorm._extra_state": ("Tensor", [0]),
        "module.decoder.layers.0.mlp.router.weight": ("Tensor", [8, 4096]),
        "module.decoder.layers.0.mlp.experts.local_experts.0.linear_fc1.weight": ("Tensor", [14336, 4096]),
        "module.decoder.layers.0.mlp.experts.local_experts.0.linear_fc1._extra_state": ("Tensor", [4]),
        "module.decoder.layers.0.mlp.experts.local_experts.0.linear_fc1.adapter.linear_in.weight": (
            "Tensor",
            [16, 4096],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.0.linear_fc1.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.0.linear_fc1.adapter.linear_out.weight": (
            "Tensor",
            [14336, 32],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.0.linear_fc1.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.0.linear_fc2.weight": ("Tensor", [4096, 7168]),
        "module.decoder.layers.0.mlp.experts.local_experts.0.linear_fc2._extra_state": ("Tensor", [4]),
        "module.decoder.layers.0.mlp.experts.local_experts.0.linear_fc2.adapter.linear_in.weight": (
            "Tensor",
            [32, 7168],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.0.linear_fc2.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.0.linear_fc2.adapter.linear_out.weight": (
            "Tensor",
            [2048, 32],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.0.linear_fc2.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.1.linear_fc1.weight": ("Tensor", [14336, 4096]),
        "module.decoder.layers.0.mlp.experts.local_experts.1.linear_fc1._extra_state": ("Tensor", [4]),
        "module.decoder.layers.0.mlp.experts.local_experts.1.linear_fc1.adapter.linear_in.weight": (
            "Tensor",
            [16, 4096],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.1.linear_fc1.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.1.linear_fc1.adapter.linear_out.weight": (
            "Tensor",
            [14336, 32],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.1.linear_fc1.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.1.linear_fc2.weight": ("Tensor", [4096, 7168]),
        "module.decoder.layers.0.mlp.experts.local_experts.1.linear_fc2._extra_state": ("Tensor", [4]),
        "module.decoder.layers.0.mlp.experts.local_experts.1.linear_fc2.adapter.linear_in.weight": (
            "Tensor",
            [32, 7168],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.1.linear_fc2.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.1.linear_fc2.adapter.linear_out.weight": (
            "Tensor",
            [2048, 32],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.1.linear_fc2.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.2.linear_fc1.weight": ("Tensor", [14336, 4096]),
        "module.decoder.layers.0.mlp.experts.local_experts.2.linear_fc1._extra_state": ("Tensor", [4]),
        "module.decoder.layers.0.mlp.experts.local_experts.2.linear_fc1.adapter.linear_in.weight": (
            "Tensor",
            [16, 4096],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.2.linear_fc1.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.2.linear_fc1.adapter.linear_out.weight": (
            "Tensor",
            [14336, 32],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.2.linear_fc1.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.2.linear_fc2.weight": ("Tensor", [4096, 7168]),
        "module.decoder.layers.0.mlp.experts.local_experts.2.linear_fc2._extra_state": ("Tensor", [4]),
        "module.decoder.layers.0.mlp.experts.local_experts.2.linear_fc2.adapter.linear_in.weight": (
            "Tensor",
            [32, 7168],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.2.linear_fc2.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.2.linear_fc2.adapter.linear_out.weight": (
            "Tensor",
            [2048, 32],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.2.linear_fc2.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.3.linear_fc1.weight": ("Tensor", [14336, 4096]),
        "module.decoder.layers.0.mlp.experts.local_experts.3.linear_fc1._extra_state": ("Tensor", [4]),
        "module.decoder.layers.0.mlp.experts.local_experts.3.linear_fc1.adapter.linear_in.weight": (
            "Tensor",
            [16, 4096],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.3.linear_fc1.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.3.linear_fc1.adapter.linear_out.weight": (
            "Tensor",
            [14336, 32],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.3.linear_fc1.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.3.linear_fc2.weight": ("Tensor", [4096, 7168]),
        "module.decoder.layers.0.mlp.experts.local_experts.3.linear_fc2._extra_state": ("Tensor", [4]),
        "module.decoder.layers.0.mlp.experts.local_experts.3.linear_fc2.adapter.linear_in.weight": (
            "Tensor",
            [32, 7168],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.3.linear_fc2.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.3.linear_fc2.adapter.linear_out.weight": (
            "Tensor",
            [2048, 32],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.3.linear_fc2.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.4.linear_fc1.weight": ("Tensor", [14336, 4096]),
        "module.decoder.layers.0.mlp.experts.local_experts.4.linear_fc1._extra_state": ("Tensor", [4]),
        "module.decoder.layers.0.mlp.experts.local_experts.4.linear_fc1.adapter.linear_in.weight": (
            "Tensor",
            [16, 4096],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.4.linear_fc1.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.4.linear_fc1.adapter.linear_out.weight": (
            "Tensor",
            [14336, 32],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.4.linear_fc1.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.4.linear_fc2.weight": ("Tensor", [4096, 7168]),
        "module.decoder.layers.0.mlp.experts.local_experts.4.linear_fc2._extra_state": ("Tensor", [4]),
        "module.decoder.layers.0.mlp.experts.local_experts.4.linear_fc2.adapter.linear_in.weight": (
            "Tensor",
            [32, 7168],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.4.linear_fc2.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.4.linear_fc2.adapter.linear_out.weight": (
            "Tensor",
            [2048, 32],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.4.linear_fc2.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.5.linear_fc1.weight": ("Tensor", [14336, 4096]),
        "module.decoder.layers.0.mlp.experts.local_experts.5.linear_fc1._extra_state": ("Tensor", [4]),
        "module.decoder.layers.0.mlp.experts.local_experts.5.linear_fc1.adapter.linear_in.weight": (
            "Tensor",
            [16, 4096],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.5.linear_fc1.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.5.linear_fc1.adapter.linear_out.weight": (
            "Tensor",
            [14336, 32],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.5.linear_fc1.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.5.linear_fc2.weight": ("Tensor", [4096, 7168]),
        "module.decoder.layers.0.mlp.experts.local_experts.5.linear_fc2._extra_state": ("Tensor", [4]),
        "module.decoder.layers.0.mlp.experts.local_experts.5.linear_fc2.adapter.linear_in.weight": (
            "Tensor",
            [32, 7168],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.5.linear_fc2.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.5.linear_fc2.adapter.linear_out.weight": (
            "Tensor",
            [2048, 32],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.5.linear_fc2.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.6.linear_fc1.weight": ("Tensor", [14336, 4096]),
        "module.decoder.layers.0.mlp.experts.local_experts.6.linear_fc1._extra_state": ("Tensor", [4]),
        "module.decoder.layers.0.mlp.experts.local_experts.6.linear_fc1.adapter.linear_in.weight": (
            "Tensor",
            [16, 4096],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.6.linear_fc1.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.6.linear_fc1.adapter.linear_out.weight": (
            "Tensor",
            [14336, 32],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.6.linear_fc1.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.6.linear_fc2.weight": ("Tensor", [4096, 7168]),
        "module.decoder.layers.0.mlp.experts.local_experts.6.linear_fc2._extra_state": ("Tensor", [4]),
        "module.decoder.layers.0.mlp.experts.local_experts.6.linear_fc2.adapter.linear_in.weight": (
            "Tensor",
            [32, 7168],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.6.linear_fc2.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.6.linear_fc2.adapter.linear_out.weight": (
            "Tensor",
            [2048, 32],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.6.linear_fc2.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.7.linear_fc1.weight": ("Tensor", [14336, 4096]),
        "module.decoder.layers.0.mlp.experts.local_experts.7.linear_fc1._extra_state": ("Tensor", [4]),
        "module.decoder.layers.0.mlp.experts.local_experts.7.linear_fc1.adapter.linear_in.weight": (
            "Tensor",
            [16, 4096],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.7.linear_fc1.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.7.linear_fc1.adapter.linear_out.weight": (
            "Tensor",
            [14336, 32],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.7.linear_fc1.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.7.linear_fc2.weight": ("Tensor", [4096, 7168]),
        "module.decoder.layers.0.mlp.experts.local_experts.7.linear_fc2._extra_state": ("Tensor", [4]),
        "module.decoder.layers.0.mlp.experts.local_experts.7.linear_fc2.adapter.linear_in.weight": (
            "Tensor",
            [32, 7168],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.7.linear_fc2.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.7.linear_fc2.adapter.linear_out.weight": (
            "Tensor",
            [2048, 32],
        ),
        "module.decoder.layers.0.mlp.experts.local_experts.7.linear_fc2.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.self_attention.core_attention._extra_state": ("Tensor", [4]),
        "module.decoder.layers.1.self_attention.linear_proj.weight": ("Tensor", [4096, 2048]),
        "module.decoder.layers.1.self_attention.linear_proj._extra_state": ("Tensor", [4]),
        "module.decoder.layers.1.self_attention.linear_proj.adapter.linear_in.weight": ("Tensor", [32, 2048]),
        "module.decoder.layers.1.self_attention.linear_proj.adapter.linear_in._extra_state": ("NoneType", []),
        "module.decoder.layers.1.self_attention.linear_proj.adapter.linear_out.weight": ("Tensor", [2048, 32]),
        "module.decoder.layers.1.self_attention.linear_proj.adapter.linear_out._extra_state": ("NoneType", []),
        "module.decoder.layers.1.self_attention.linear_qkv.layer_norm_weight": ("Tensor", [4096]),
        "module.decoder.layers.1.self_attention.linear_qkv.weight": ("Tensor", [3072, 4096]),
        "module.decoder.layers.1.self_attention.linear_qkv._extra_state": ("Tensor", [4]),
        "module.decoder.layers.1.self_attention.linear_qkv.adapter.linear_in.weight": ("Tensor", [16, 4096]),
        "module.decoder.layers.1.self_attention.linear_qkv.adapter.linear_in._extra_state": ("NoneType", []),
        "module.decoder.layers.1.self_attention.linear_qkv.adapter.linear_out.weight": ("Tensor", [3072, 32]),
        "module.decoder.layers.1.self_attention.linear_qkv.adapter.linear_out._extra_state": ("NoneType", []),
        "module.decoder.layers.1.pre_mlp_layernorm.weight": ("Tensor", [4096]),
        "module.decoder.layers.1.pre_mlp_layernorm._extra_state": ("Tensor", [0]),
        "module.decoder.layers.1.mlp.router.weight": ("Tensor", [8, 4096]),
        "module.decoder.layers.1.mlp.experts.local_experts.0.linear_fc1.weight": ("Tensor", [14336, 4096]),
        "module.decoder.layers.1.mlp.experts.local_experts.0.linear_fc1._extra_state": ("Tensor", [4]),
        "module.decoder.layers.1.mlp.experts.local_experts.0.linear_fc1.adapter.linear_in.weight": (
            "Tensor",
            [16, 4096],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.0.linear_fc1.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.0.linear_fc1.adapter.linear_out.weight": (
            "Tensor",
            [14336, 32],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.0.linear_fc1.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.0.linear_fc2.weight": ("Tensor", [4096, 7168]),
        "module.decoder.layers.1.mlp.experts.local_experts.0.linear_fc2._extra_state": ("Tensor", [4]),
        "module.decoder.layers.1.mlp.experts.local_experts.0.linear_fc2.adapter.linear_in.weight": (
            "Tensor",
            [32, 7168],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.0.linear_fc2.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.0.linear_fc2.adapter.linear_out.weight": (
            "Tensor",
            [2048, 32],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.0.linear_fc2.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.1.linear_fc1.weight": ("Tensor", [14336, 4096]),
        "module.decoder.layers.1.mlp.experts.local_experts.1.linear_fc1._extra_state": ("Tensor", [4]),
        "module.decoder.layers.1.mlp.experts.local_experts.1.linear_fc1.adapter.linear_in.weight": (
            "Tensor",
            [16, 4096],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.1.linear_fc1.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.1.linear_fc1.adapter.linear_out.weight": (
            "Tensor",
            [14336, 32],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.1.linear_fc1.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.1.linear_fc2.weight": ("Tensor", [4096, 7168]),
        "module.decoder.layers.1.mlp.experts.local_experts.1.linear_fc2._extra_state": ("Tensor", [4]),
        "module.decoder.layers.1.mlp.experts.local_experts.1.linear_fc2.adapter.linear_in.weight": (
            "Tensor",
            [32, 7168],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.1.linear_fc2.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.1.linear_fc2.adapter.linear_out.weight": (
            "Tensor",
            [2048, 32],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.1.linear_fc2.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.2.linear_fc1.weight": ("Tensor", [14336, 4096]),
        "module.decoder.layers.1.mlp.experts.local_experts.2.linear_fc1._extra_state": ("Tensor", [4]),
        "module.decoder.layers.1.mlp.experts.local_experts.2.linear_fc1.adapter.linear_in.weight": (
            "Tensor",
            [16, 4096],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.2.linear_fc1.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.2.linear_fc1.adapter.linear_out.weight": (
            "Tensor",
            [14336, 32],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.2.linear_fc1.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.2.linear_fc2.weight": ("Tensor", [4096, 7168]),
        "module.decoder.layers.1.mlp.experts.local_experts.2.linear_fc2._extra_state": ("Tensor", [4]),
        "module.decoder.layers.1.mlp.experts.local_experts.2.linear_fc2.adapter.linear_in.weight": (
            "Tensor",
            [32, 7168],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.2.linear_fc2.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.2.linear_fc2.adapter.linear_out.weight": (
            "Tensor",
            [2048, 32],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.2.linear_fc2.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.3.linear_fc1.weight": ("Tensor", [14336, 4096]),
        "module.decoder.layers.1.mlp.experts.local_experts.3.linear_fc1._extra_state": ("Tensor", [4]),
        "module.decoder.layers.1.mlp.experts.local_experts.3.linear_fc1.adapter.linear_in.weight": (
            "Tensor",
            [16, 4096],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.3.linear_fc1.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.3.linear_fc1.adapter.linear_out.weight": (
            "Tensor",
            [14336, 32],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.3.linear_fc1.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.3.linear_fc2.weight": ("Tensor", [4096, 7168]),
        "module.decoder.layers.1.mlp.experts.local_experts.3.linear_fc2._extra_state": ("Tensor", [4]),
        "module.decoder.layers.1.mlp.experts.local_experts.3.linear_fc2.adapter.linear_in.weight": (
            "Tensor",
            [32, 7168],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.3.linear_fc2.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.3.linear_fc2.adapter.linear_out.weight": (
            "Tensor",
            [2048, 32],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.3.linear_fc2.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.4.linear_fc1.weight": ("Tensor", [14336, 4096]),
        "module.decoder.layers.1.mlp.experts.local_experts.4.linear_fc1._extra_state": ("Tensor", [4]),
        "module.decoder.layers.1.mlp.experts.local_experts.4.linear_fc1.adapter.linear_in.weight": (
            "Tensor",
            [16, 4096],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.4.linear_fc1.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.4.linear_fc1.adapter.linear_out.weight": (
            "Tensor",
            [14336, 32],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.4.linear_fc1.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.4.linear_fc2.weight": ("Tensor", [4096, 7168]),
        "module.decoder.layers.1.mlp.experts.local_experts.4.linear_fc2._extra_state": ("Tensor", [4]),
        "module.decoder.layers.1.mlp.experts.local_experts.4.linear_fc2.adapter.linear_in.weight": (
            "Tensor",
            [32, 7168],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.4.linear_fc2.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.4.linear_fc2.adapter.linear_out.weight": (
            "Tensor",
            [2048, 32],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.4.linear_fc2.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.5.linear_fc1.weight": ("Tensor", [14336, 4096]),
        "module.decoder.layers.1.mlp.experts.local_experts.5.linear_fc1._extra_state": ("Tensor", [4]),
        "module.decoder.layers.1.mlp.experts.local_experts.5.linear_fc1.adapter.linear_in.weight": (
            "Tensor",
            [16, 4096],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.5.linear_fc1.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.5.linear_fc1.adapter.linear_out.weight": (
            "Tensor",
            [14336, 32],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.5.linear_fc1.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.5.linear_fc2.weight": ("Tensor", [4096, 7168]),
        "module.decoder.layers.1.mlp.experts.local_experts.5.linear_fc2._extra_state": ("Tensor", [4]),
        "module.decoder.layers.1.mlp.experts.local_experts.5.linear_fc2.adapter.linear_in.weight": (
            "Tensor",
            [32, 7168],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.5.linear_fc2.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.5.linear_fc2.adapter.linear_out.weight": (
            "Tensor",
            [2048, 32],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.5.linear_fc2.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.6.linear_fc1.weight": ("Tensor", [14336, 4096]),
        "module.decoder.layers.1.mlp.experts.local_experts.6.linear_fc1._extra_state": ("Tensor", [4]),
        "module.decoder.layers.1.mlp.experts.local_experts.6.linear_fc1.adapter.linear_in.weight": (
            "Tensor",
            [16, 4096],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.6.linear_fc1.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.6.linear_fc1.adapter.linear_out.weight": (
            "Tensor",
            [14336, 32],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.6.linear_fc1.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.6.linear_fc2.weight": ("Tensor", [4096, 7168]),
        "module.decoder.layers.1.mlp.experts.local_experts.6.linear_fc2._extra_state": ("Tensor", [4]),
        "module.decoder.layers.1.mlp.experts.local_experts.6.linear_fc2.adapter.linear_in.weight": (
            "Tensor",
            [32, 7168],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.6.linear_fc2.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.6.linear_fc2.adapter.linear_out.weight": (
            "Tensor",
            [2048, 32],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.6.linear_fc2.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.7.linear_fc1.weight": ("Tensor", [14336, 4096]),
        "module.decoder.layers.1.mlp.experts.local_experts.7.linear_fc1._extra_state": ("Tensor", [4]),
        "module.decoder.layers.1.mlp.experts.local_experts.7.linear_fc1.adapter.linear_in.weight": (
            "Tensor",
            [16, 4096],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.7.linear_fc1.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.7.linear_fc1.adapter.linear_out.weight": (
            "Tensor",
            [14336, 32],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.7.linear_fc1.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.7.linear_fc2.weight": ("Tensor", [4096, 7168]),
        "module.decoder.layers.1.mlp.experts.local_experts.7.linear_fc2._extra_state": ("Tensor", [4]),
        "module.decoder.layers.1.mlp.experts.local_experts.7.linear_fc2.adapter.linear_in.weight": (
            "Tensor",
            [32, 7168],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.7.linear_fc2.adapter.linear_in._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.7.linear_fc2.adapter.linear_out.weight": (
            "Tensor",
            [2048, 32],
        ),
        "module.decoder.layers.1.mlp.experts.local_experts.7.linear_fc2.adapter.linear_out._extra_state": (
            "NoneType",
            [],
        ),
        "module.decoder.final_layernorm.weight": ("Tensor", [4096]),
        "module.decoder.final_layernorm._extra_state": ("Tensor", [0]),
        "module.output_layer.weight": ("Tensor", [15104, 4096]),
        "module.output_layer._extra_state": ("NoneType", []),
    }


def mixtral_8x7b() -> pl.LightningModule:
    tokenizer = OrdTokenizer()
    model = llm.MixtralModel(llm.MixtralConfig8x7B(num_layers=2), tokenizer=tokenizer)
    return model, get_mixtral_expected_ckpt()


def mistral_7b() -> pl.LightningModule:
    tokenizer = OrdTokenizer()
    model = llm.MistralModel(llm.MistralConfig7B(num_layers=2), tokenizer=tokenizer)
    return model, get_mistral_expected_ckpt()


def make_lora(use_exclude):
    if use_exclude:
        return llm.peft.LoRA(exclude_modules=['output_layer'], target_modules=[])
    else:
        return llm.peft.LoRA()


if __name__ == '__main__':
    args = get_args()
    if args.model == 'mistral':
        model, ref_ckpt = mistral_7b()
    else:
        model, ref_ckpt = mixtral_8x7b()

    lora = make_lora(args.use_exclude)

    llm.finetune(
        model=model,
        data=squad(args.mbs, args.gbs),
        trainer=trainer(args.tp, args.tp, args.ep, args.tp > 1, args.max_steps),
        peft=lora,
        log=logger(),
        optim=nl.MegatronOptimizerModule(
            config=OptimizerConfig(
                optimizer="adam",
                lr=0.0001,
                adam_beta2=0.98,
                use_distributed_optimizer=args.dist_opt,
                clip_grad=1.0,
                bf16=True,
            ),
        ),
    )
    if args.tp == 2:
        for key, val in model.state_dict().items():
            if isinstance(val, torch.Tensor):
                shape = list(val.size())
            else:
                shape = []
            if not '_extra_state' in key:
                assert ref_ckpt[key] == (f"{type(val).__name__}", shape)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
from dataclasses import dataclass
from typing import Callable, Dict, Literal, Optional

import lightning.pytorch as L
import torch
import torch.nn.functional as F
from megatron.core import parallel_state
from megatron.core.utils import get_batch_on_this_cp_rank
from torch import Tensor, nn

from nemo.collections.common.tokenizers import TokenizerSpec
from nemo.collections.llm.bert.loss import BERTInBatchExclusiveHardNegativesRankingLoss
from nemo.collections.llm.bert.model import BertConfig, BertModel
from nemo.collections.llm.bert.model.base import get_packed_seq_params
from nemo.collections.llm.bert.model.bert import HuggingFaceBertImporter
from nemo.lightning import io
from nemo.lightning.pytorch.optim import OptimizerModule


def bert_embedding_data_step(dataloder_iter) -> Dict[str, torch.Tensor]:
    """Setup BERT dataloader batch."""
    batch = next(dataloder_iter)

    _batch: dict
    if isinstance(batch, tuple) and len(batch) == 3:
        _batch = batch[0]
    else:
        _batch = batch

    required_keys = set()
    required_keys.add("attention_mask")
    required_keys.add("token_type_ids")

    if parallel_state.is_pipeline_first_stage():
        required_keys.add("input_ids")

    _batch = {key: val.cuda(non_blocking=True) if key in required_keys else None for key, val in _batch.items()}
    # slice batch along sequence dimension for context parallelism
    output = get_batch_on_this_cp_rank(_batch)

    return output


def bert_embedding_forward_step(model: L.LightningModule, batch: Dict[str, torch.Tensor]) -> torch.Tensor:
    """
    This subsets the batch keys to the ones actually used by forward pass of the model,
    and then calls the model's forward pass. if "cu_seqsens" are defined in the batch,
    then the packed sequence parameters are also passed to the model for forward pass efficiency.
    """
    forward_args = {
        "input_ids": batch["input_ids"],
        "attention_mask": batch["attention_mask"],
    }

    if model.config.num_tokentypes != 0:
        forward_args["tokentype_ids"] = batch["token_type_ids"]

    if "cu_seqlens" in batch:
        forward_args["packed_seq_params"] = get_packed_seq_params(batch)

    return model(**forward_args)


@dataclass
class BertEmbeddingConfig(BertConfig):
    """Bert Embedding Config"""

    bert_type: Literal["huggingface", "megatron"] = 'huggingface'
    ce_loss_scale: float = 20
    label_smoothing: float = 0.0
    add_lm_head: bool = False
    bert_binary_head: bool = False
    num_hard_negatives: int = 1
    num_tokentypes: int = 2
    global_in_batch_negatives: bool = True
    backprop_type: Literal["local", "global"] = 'local'
    forward_step_fn: Callable = bert_embedding_forward_step
    data_step_fn: Callable = bert_embedding_data_step


@dataclass
class BertEmbeddingLargeConfig(BertEmbeddingConfig):
    """Bert Embedding model follows Bert-large architecture."""

    num_layers: int = 24
    hidden_size: int = 1024
    intermediate_size: int = 4096
    num_attention_heads: int = 16


@dataclass
class BertEmbeddingMiniConfig(BertEmbeddingConfig):
    """Bert Embedding model follows Bert-mini (384 hidden size) architecture."""

    num_layers: int = 6
    hidden_size: int = 384
    intermediate_size: int = 1536
    num_attention_heads: int = 12


class BertEmbeddingHead(nn.Module):
    """Performs mean pooling on the token embeddings."""

    def __init__(
        self,
        word_embedding_dimension: int,
        pooling_mode_mean_tokens: bool = True,
    ):
        super(BertEmbeddingHead, self).__init__()

        self.config_keys = [
            "word_embedding_dimension",
            "pooling_mode_mean_tokens",
        ]
        self.word_embedding_dimension = word_embedding_dimension
        self.pooling_mode_mean_tokens = pooling_mode_mean_tokens

    def forward(self, token_embeddings: Tensor, attention_mask: Tensor):
        """Forward function for embedding head. Performs mean pooling."""
        token_embeddings = token_embeddings.permute(1, 0, 2)
        input_mask_expanded = attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
        sum_embeddings = torch.sum(token_embeddings * input_mask_expanded, 1)

        sum_mask = input_mask_expanded.sum(1)
        sum_mask = torch.clamp(sum_mask, min=1e-9)

        output_vector = sum_embeddings / sum_mask
        output_vector = F.normalize(output_vector, p=2, dim=1)

        return output_vector


class BertEmbeddingModel(BertModel):
    """Bert Lightning Module"""

    def __init__(
        self,
        config: BertConfig,
        # TODO: Add transformer_layer_spec when we update mcore
        optim: Optional[OptimizerModule] = None,
        tokenizer: Optional["TokenizerSpec"] = None,
        model_transform: Optional[Callable[[nn.Module], nn.Module]] = None,
    ):
        super().__init__(config, optim, tokenizer, model_transform)

    def configure_model(self) -> None:
        """Setup the BERT Model based on config definition."""
        if not hasattr(self, "module"):
            self.module = self.config.configure_model(self.tokenizer)
            self.embedding_head = BertEmbeddingHead(
                word_embedding_dimension=self.config.hidden_size,
                pooling_mode_mean_tokens=True,
            )

    def forward(
        self,
        *args,
        **kwargs,
    ) -> torch.Tensor:
        """Call the forward method of the underlying model, and return whatever it outputs."""
        assert "attention_mask" in kwargs, "attention mask is required for BERT Embedding Model."
        output_tensor = self.module(
            hidden_states_only=True, *args, **kwargs
        )  # for now just pass through to the underlying model
        embeddings_out = self.embedding_head(output_tensor, kwargs["attention_mask"])
        return embeddings_out

    @property
    def training_loss_reduction(self) -> BERTInBatchExclusiveHardNegativesRankingLoss:  # pylint: disable=C0115,C0116
        if not self._training_loss_reduction:
            self._training_loss_reduction = BERTInBatchExclusiveHardNegativesRankingLoss(
                validation_step=False,
                num_hard_negatives=self.config.num_hard_negatives,
                scale=self.config.ce_loss_scale,
                label_smoothing=self.config.label_smoothing,
                global_in_batch_negatives=self.config.global_in_batch_negatives,
                backprop_type=self.config.backprop_type,
            )

        return self._training_loss_reduction

    @property
    def validation_loss_reduction(self) -> BERTInBatchExclusiveHardNegativesRankingLoss:  # pylint: disable=C0115,C0116
        if not self._validation_loss_reduction:
            self._validation_loss_reduction = BERTInBatchExclusiveHardNegativesRankingLoss(
                validation_step=True,
                num_hard_negatives=self.config.num_hard_negatives,
                scale=self.config.ce_loss_scale,
                label_smoothing=self.config.label_smoothing,
            )

        return self._validation_loss_reduction


@io.model_importer(BertEmbeddingModel, "hf")
class BertEmbeddingImporter(HuggingFaceBertImporter):
    """
    Importer for BertEmbedding Model.
    HuggingFace uses same model for Bert Embedding model and Bert model, thus the connector is identical.
    """

    def __init__(self, *args, **kwargs):
        if sys.version_info > (3, 11):
            # In Python versions <= 3.11, *Path classes don’t have a __init__ method,
            # and do all their initialization in __new__/ helper methods.
            # Only need to call super().__init__ if version > 3.11
            super().__init__(*args)
        self.type = 'model'

    def init(self) -> BertEmbeddingModel:
        return BertEmbeddingModel(self.config, tokenizer=self.tokenizer)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os.path
from io import BytesIO
from pathlib import Path
from typing import Any, Dict, Union

import numpy

# tenosrstore is needed to register 'bfloat16' dtype with numpy for zarr compatibility
import tensorstore  # noqa: F401 pylint: disable=unused-import
import torch
from torch.distributed.checkpoint import FileSystemReader, load
from torch.distributed.checkpoint.metadata import BytesStorageMetadata, TensorStorageMetadata

from nemo.export.tarutils import TarPath, ZarrPathStore
from nemo.export.utils._mock_import import _mock_import

LOGGER = logging.getLogger("NeMo")


def nemo_to_path(nemo_checkpoint: Union[Path, str]) -> Union[Path, TarPath]:
    """
    Creates Path / TarPath object suitable for navigating inside the nemo checkpoint.

    Args:
        nemo_checkpoint (Path, str): Path to the NeMo checkpoint.
    Returns:
        Path | TarPath: Suitable Path object for navigating through the checkpoint.
    """
    string_path = str(nemo_checkpoint)

    if os.path.isdir(string_path):
        return Path(string_path)
    return TarPath(string_path)


class TarFileSystemReader(FileSystemReader):
    """Reader that accepts both Path and TarPath checkpoint directory.

    The FileSystemReader works with TarPath, but expects a pure Path.
    It's enough to skip the Path check in __init__.
    """

    def __init__(self, path: Union[Path, TarPath]) -> None:
        """Makes sure that super().__init__ gets a pure path as expected."""
        super_path = str(path) if isinstance(path, TarPath) else path
        super().__init__(super_path)
        if isinstance(path, TarPath):
            self.path = path  # overwrites path set in super().__init__ call


def load_sharded_metadata_torch_dist(
    checkpoint_dir: Union[Path, TarPath], load_extra_states: bool = False
) -> Dict[str, Any]:
    """
    Loads model state dictionary from torch_dist checkpoint.

    Args:
        checkpoint_dir (Path | TarPath): Path to the model weights directory.
        load_extra_states (bool): If set to true, loads BytesIO objects, related to the extra states.
    Returns:
        dict: Loaded model state dictionary (weights are stored in torch tensors).
    """
    fs_reader = TarFileSystemReader(checkpoint_dir)
    metadata = fs_reader.read_metadata()

    state_dict = {
        k: torch.empty(tp.size, dtype=tp.properties.dtype)
        for k, tp in metadata.state_dict_metadata.items()
        if isinstance(tp, TensorStorageMetadata)
    }

    if load_extra_states:
        state_dict.update(
            {k: [] for k, tp in metadata.state_dict_metadata.items() if isinstance(tp, BytesStorageMetadata)}
        )

    load(state_dict, storage_reader=fs_reader)
    return state_dict


def load_sharded_pickle_extra_state_scale(dir: Union[Path, TarPath]) -> Dict[str, BytesIO]:
    """
    Loads model extra states from the .pt shards.

    Args:
        dir (Path | TarPath): Path to the directory with sharded extra states.
    Returns:
        dict: State dictionary corresponding to the loaded extra states.
    """
    pt_files = list(dir.glob('shard_*_*.pt'))
    extra_states = {}
    for file in pt_files:
        shard_name = file.name.split('.')[0]
        with file.open('rb') as opened_file:
            extra_states[dir.name + '/' + shard_name] = torch.load(opened_file, weights_only=True)

    return extra_states


def contains_extra_states(subdir: Union[Path, TarPath]) -> bool:
    """
    Checks if zarr directory contains extra states.

    Args:
        subdir (Path | TarPath): Directory inside the zarr checkpoint.
    Returns:
        bool: Is a directory with extra states
    """
    return list(subdir.glob('shard_0_*.pt')) != []


def load_sharded_metadata_zarr(
    checkpoint_dir: Union[Path, TarPath], load_extra_states: bool = False
) -> Dict[str, Any]:
    """
    Loads model dictionary from the zarr format.

    Args:
        checkpoint_dir (Path | TarPath): Path to the NeMo checkpoint.
        load_extra_states (bool): If set to True, the function will load BufferIO objects with extra states.
    Returns:
        dict: Model state dictionary.
    """
    if load_extra_states:
        torch.serialization.add_safe_globals([BytesIO])

    sharded_state_dict = {}
    for subdir in checkpoint_dir.iterdir():
        if not subdir.is_dir():
            continue

        if load_extra_states and contains_extra_states(subdir):
            sharded_state_dict.update(load_sharded_pickle_extra_state_scale(subdir))

        elif (subdir / '.zarray').exists():
            key = subdir.name
            zstore = ZarrPathStore(subdir)

            import zarr

            arr = zarr.open(zstore, 'r')

            if arr.dtype.name == "bfloat16":
                sharded_state_dict[key] = torch.from_numpy(arr[:].view(numpy.int16)).view(torch.bfloat16)
            else:
                sharded_state_dict[key] = torch.from_numpy(arr[:])

    return sharded_state_dict


def nemo_weights_directory(nemo_path: Union[Path, TarPath]) -> Union[Path, TarPath]:
    """
    Returns a Path pointing to the weights directory inside the NeMo checkpoint.

    Args:
        nemo_path (Path | TarPath): Path to the nemo checkpoint.
    Returns:
        Path | TarPath: Path to the weights directory inside the model checkpoint.
    """
    if (nemo_path / "model_weights").exists():
        return nemo_path / "model_weights"

    if (nemo_path / "weights").exists():
        return nemo_path / "weights"

    return nemo_path


def load_model_weights(checkpoint_path: Union[str, Path], load_extra_states: bool = False) -> Dict[str, Any]:
    """
    Loads NeMo state dictionary. Weights are stored in torch.Tensor

    Args:
        checkpoint_path (str | Path): Path to the NeMo checkpoint.
        load_extra_states (bool): If True, loads BytesIO objects, corresponding to the extra states.
    Returns:
        dict: Model state dictionary.
    """

    nemo_path = nemo_to_path(checkpoint_path)
    nemo_weights = nemo_weights_directory(nemo_path)

    with (nemo_weights / 'metadata.json').open(mode='r') as f:
        config_dict = json.load(f)

    if config_dict['sharded_backend'] == 'zarr':
        return load_sharded_metadata_zarr(nemo_weights, load_extra_states=load_extra_states)
    elif config_dict['sharded_backend'] == 'torch_dist':
        # TODO: Remove mocking imports once MCore is available in NIM containers
        with _mock_import("megatron.core.dist_checkpointing.strategies.torch"):
            return load_sharded_metadata_torch_dist(nemo_weights, load_extra_states=load_extra_states)

    raise NotImplementedError(f'Distributed checkpoint backend {config_dict["sharded_backend"]} not supported')



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test the LLaMA3 recipe with a smaller model.
"""

import argparse
import os

import nemo_run as run
import torch

from nemo.collections import llm
from nemo.lightning.pytorch.callbacks.debugging import ParameterDebugger
from nemo.lightning.pytorch.callbacks.pytorch_profiler import PytorchProfilerCallback
from tests.collections.llm.common import (
    AssertOptimizerParamGroupsHaveAtLeastTwoWeightDecays,
    MCoreModelAttributeValidator,
    MiscAttributeValidator,
    StopBeforeEnd,
    create_verify_precision,
    small_llama_cfg,
    train_data,
    verify_ckpt_dir,
)


def get_args():
    parser = argparse.ArgumentParser(prog="", description="")
    parser.add_argument('--devices', type=int, required=True, help="Number of devices to use for training")
    parser.add_argument('--max-steps', type=int, required=True, help="Number of steps to train for")
    parser.add_argument(
        '--early-stop',
        type=int,
        default=None,
        help="Stop training early at this global step (for testing resume training)",
    )
    parser.add_argument(
        '--experiment-dir', type=str, required=True, help="directory to write results and checkpoints to"
    )
    parser.add_argument(
        '--data-path', type=str, default=None, help="Path to data file. If not specified, uses mock data."
    )
    parser.add_argument(
        '--tokenizer-path',
        type=str,
        default=None,
        help="Path to a sentencepiece tokenizer model file. If not specified, uses mock data.",
    )
    parser.add_argument('--index-mapping-dir', type=str, help="directory to write index mappings to")
    parser.add_argument('--seq-length', type=int, default=8192, help="Sequence length. default is 8k")
    parser.add_argument('--tp', type=int, default=None, help="Override tensor parallelism")
    parser.add_argument('--pp', type=int, default=None, help="Override pipeline parallelism")
    parser.add_argument('--vp', type=int, default=None, help="Override virtual pipeline parallelism")
    parser.add_argument('--cp', type=int, default=None, help="Override context parallelism")
    parser.add_argument('--sp', type=int, choices=[0, 1], default=None, help="Override sequence parallel")
    parser.add_argument(
        '--precision', type=str, choices=['bf16', 'fp16', 'fp32'], default='bf16', help="Override recipe precision"
    )
    parser.add_argument('--fp8', action='store_true', help="Enable FP8")
    parser.add_argument(
        '--profiler',
        action='store_true',
        help="Attach PytorchProfilerCallback and verify trace files after training",
    )
    parser.add_argument(
        '--ckpt-optim-fully-reshardable', action='store_true', help="Enable optimizer checkpoint fully-reshardability"
    )

    return parser.parse_args()


def main():
    args = get_args()

    exp_name = "L2_llama3_small_pretrain_test"
    pretrain_recipe = llm.llama3_8b.pretrain_recipe(
        dir=args.experiment_dir, name=exp_name, num_gpus_per_node=args.devices
    )

    pretrain_recipe.model = run.Config(llm.LlamaModel, small_llama_cfg(args.seq_length))

    if args.data_path and args.tokenizer_path:
        pretrain_recipe.data = train_data(
            data_path=args.data_path,
            tokenizer_path=args.tokenizer_path,
            index_mapping_dir=args.index_mapping_dir,
            seq_length=args.seq_length,
        )

    # Recipe Overrides
    pretrain_recipe.trainer.max_steps = args.max_steps
    pretrain_recipe.trainer.log_every_n_steps = 1
    pretrain_recipe.log.ckpt.every_n_train_steps = None
    pretrain_recipe.log.ckpt.train_time_interval = None
    pretrain_recipe.trainer.val_check_interval = 2
    pretrain_recipe.trainer.limit_val_batches = 2

    if args.early_stop:
        pretrain_recipe.trainer.callbacks.append(StopBeforeEnd(stop_on_step=args.early_stop))
    pretrain_recipe.trainer.callbacks.append(AssertOptimizerParamGroupsHaveAtLeastTwoWeightDecays())

    if args.ckpt_optim_fully_reshardable:
        pretrain_recipe.trainer.strategy.ckpt_optim_fully_reshardable = True

    if not args.precision == 'bf16' or args.fp8:  # default case is bf16 without fp8
        import llm.recipes.precision.mixed_precision as mp_recipes

        key = (args.precision, args.fp8)
        precision_recipe = {
            ("fp16", False): mp_recipes.fp16_mixed,
            ("bf16", True): mp_recipes.bf16_with_fp8_mixed,
            ("fp16", True): mp_recipes.fp16_with_fp8_mixed,
            # Need fp32
        }[key]
        pretrain_recipe.trainer.plugins = precision_recipe()
    dtype_map = {"bf16": torch.bfloat16, "fp16": torch.float16, "fp32": torch.float32}
    debugger_callback = ParameterDebugger(
        param_fn=create_verify_precision(dtype_map[args.precision]),
        grad_fn=create_verify_precision(torch.float32),
        log_on_hooks=["on_train_start", "on_train_end"],
    )
    pretrain_recipe.trainer.callbacks.append(debugger_callback)

    parallelisms = {
        "tensor_model_parallel_size": args.tp,
        "pipeline_model_parallel_size": args.pp,
        "virtual_pipeline_model_parallel_size": args.vp,
        "context_parallel_size": args.cp,
        "sequence_parallel": bool(args.sp) if args.sp is not None else None,
    }
    for k, v in parallelisms.items():
        if v is not None:  # use recipe default if not specified
            setattr(pretrain_recipe.trainer.strategy, k, v)
        parallelisms[k] = getattr(pretrain_recipe.trainer.strategy, k)
    pretrain_recipe.trainer.callbacks.append(MCoreModelAttributeValidator(parallelisms))

    misc_checker = MiscAttributeValidator(
        {"max_steps": args.max_steps, "stop_on_step": args.early_stop or args.max_steps}
    )
    pretrain_recipe.trainer.callbacks.append(misc_checker)

    if args.profiler:
        exp_path = os.path.join(args.experiment_dir, exp_name)
        trace_dir = os.path.join(exp_path, "traces")
        os.makedirs(trace_dir, exist_ok=True)
        profiler_cb = PytorchProfilerCallback(
            start_step=0,
            end_step=args.max_steps,
            warmup_steps=0,
            active_steps=args.max_steps,
            trace_dir=trace_dir,
            profiler_kwargs={'with_stack': True},
        )
        pretrain_recipe.trainer.callbacks.append(profiler_cb)

    run.run(pretrain_recipe, direct=True)

    verify_ckpt_dir(
        pretrain_recipe.log.ckpt,
        args.early_stop or args.max_steps,
        pretrain_recipe.trainer.val_check_interval,
        os.path.join(args.experiment_dir, exp_name),
    )

    if args.profiler:
        exp_path = os.path.join(args.experiment_dir, exp_name)
        trace_root = os.path.join(exp_path, "traces")
        device_dir = os.path.join(trace_root, "device")
        host_dir = os.path.join(trace_root, "host")

        assert os.path.isdir(device_dir), f"Missing device traces directory: {device_dir}"
        assert os.path.isdir(host_dir), f"Missing host traces directory: {host_dir}"

        device_jsons = [f for f in os.listdir(device_dir) if f.endswith(".json")]
        host_jsons = [f for f in os.listdir(host_dir) if f.endswith(".json")]

        assert (
            len(device_jsons) == args.devices
        ), f"Expected {args.devices} JSON files in {device_dir}, found {len(device_jsons)}"
        assert (
            len(host_jsons) == args.devices
        ), f"Expected {args.devices} JSON files in {host_dir}, found {len(host_jsons)}"


if __name__ == '__main__':
    main()



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import torch
from torchmetrics import Metric

__all__ = ['MultiBinaryAccuracy']


class MultiBinaryAccuracy(Metric):
    """
    This metric computes accuracies that are needed to evaluate multiple binary outputs.
    For example, if a model returns a set of multiple sigmoid outputs per each sample or at each time step,
    F1 score can be calculated to monitor Type 1 error and Type 2 error together.

    Example:
        def validation_step(self, batch, batch_idx):
            ...
            signals, signal_lengths, targets = batch
            preds, _ = self.forward(input_signal=signals,
                                    signal_lengths=signal_lengths,
                                    targets=targets)
            loss = self.loss(logits=preds, labels=targets)
            self._accuracy_valid(preds, targets, signal_lengths)
            f1_acc = self._accuracy.compute()
            self.val_outputs = {'val_loss': loss, 'val_f1_acc': f1_acc}
            return self.val_outputs

        def on_validation_epoch_end(self):
            ...
            val_loss_mean = torch.stack([x['val_loss'] for x in self.val_outputs]).mean()
            correct_counts = torch.stack([x['val_correct_counts'] for x in self.val_outputs]).sum(axis=0)
            total_counts = torch.stack([x['val_total_counts'] for x in self.val_outputs]).sum(axis=0)

            self._accuracy_valid.correct_counts_k = correct_counts
            self._accuracy_valid.total_counts_k = total_counts
            f1_acc = self._accuracy_valid.compute()
            self._accuracy_valid.reset()

            self.log('val_loss', val_loss_mean)
            self.log('val_f1_acc', f1_acc)
            self.val_outputs.clear()  # free memory
            return {'val_loss': val_loss_mean, 'val_f1_acc': f1_acc}

    Args:
        preds (torch.Tensor):
            Predicted values which should be in range of [0, 1].
        targets (torch.Tensor):
            Target values which should be in range of [0, 1].
        signal_lengths (torch.Tensor):
            Length of each sequence in the batch input. signal_lengths values are used to
            filter out zero-padded parts in each sequence.

    Returns:
        f1_score (torch.Tensor):
            F1 score calculated from the predicted value and binarized target values.
    """

    full_state_update = False

    def __init__(self, dist_sync_on_step=False):
        super().__init__(dist_sync_on_step=dist_sync_on_step)
        self.add_state("total_correct_counts", default=torch.tensor(0), dist_reduce_fx='sum', persistent=False)
        self.add_state("total_sample_counts", default=torch.tensor(0), dist_reduce_fx='sum', persistent=False)
        self.add_state("true_positive_count", default=torch.tensor(0), dist_reduce_fx='sum', persistent=False)
        self.add_state("false_positive_count", default=torch.tensor(0), dist_reduce_fx='sum', persistent=False)
        self.add_state("false_negative_count", default=torch.tensor(0), dist_reduce_fx='sum', persistent=False)
        self.add_state("positive_count", default=torch.tensor(0), dist_reduce_fx='sum', persistent=False)
        self.eps = 1e-6

    @torch.inference_mode()
    def update(
        self, preds: torch.Tensor, targets: torch.Tensor, signal_lengths: torch.Tensor, cumulative=False, **kwargs
    ) -> torch.Tensor:
        """
        Update the metric with the given predictions, targets, and signal lengths to the metric instance.

        Args:
            preds (torch.Tensor): Predicted values.
            targets (torch.Tensor): Target values.
            signal_lengths (torch.Tensor): Length of each sequence in the batch input.
            cumulative (bool): Whether to accumulate the values over time.
        """
        preds_list = [preds[k, : signal_lengths[k], :] for k in range(preds.shape[0])]
        targets_list = [targets[k, : signal_lengths[k], :] for k in range(targets.shape[0])]
        self.preds = torch.cat(preds_list, dim=0)
        self.targets = torch.cat(targets_list, dim=0)

        self.true = self.preds.round().bool() == self.targets.round().bool()
        self.false = self.preds.round().bool() != self.targets.round().bool()
        self.positive = self.preds.round().bool() == 1
        self.negative = self.preds.round().bool() == 0

        if cumulative:
            self.positive_count += torch.sum(self.preds.round().bool() == True)
            self.true_positive_count += torch.sum(torch.logical_and(self.true, self.positive))
            self.false_positive_count += torch.sum(torch.logical_and(self.false, self.positive))
            self.false_negative_count += torch.sum(torch.logical_and(self.false, self.negative))
            self.total_correct_counts += torch.sum(self.preds.round().bool() == self.targets.round().bool())
            self.total_sample_counts += torch.prod(torch.tensor(self.targets.shape))
        else:
            self.positive_count = torch.sum(self.preds.round().bool() == True)
            self.true_positive_count = torch.sum(torch.logical_and(self.true, self.positive))
            self.false_positive_count = torch.sum(torch.logical_and(self.false, self.positive))
            self.false_negative_count = torch.sum(torch.logical_and(self.false, self.negative))
            self.total_correct_counts = torch.sum(self.preds.round().bool() == self.targets.round().bool())
            self.total_sample_counts = torch.prod(torch.tensor(self.targets.shape))

    def compute(self):
        """
        Compute F1 score from the accumulated values. Return -1 if the F1 score is NaN.

        Returns:
            f1_score (torch.Tensor): F1 score calculated from the accumulated values.
            precision (torch.Tensor): Precision calculated from the accumulated values.
            recall (torch.Tensor): Recall calculated from the accumulated values.
        """
        precision = self.true_positive_count / (self.true_positive_count + self.false_positive_count + self.eps)
        recall = self.true_positive_count / (self.true_positive_count + self.false_negative_count + self.eps)
        f1_score = (2 * precision * recall / (precision + recall + self.eps)).detach().clone()

        if torch.isnan(f1_score):
            logging.warning("self.f1_score contains NaN value. Returning -1 instead of NaN value.")
            f1_score = -1
        return f1_score.float(), precision.float(), recall.float()



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Blendable dataset."""

import os
import subprocess
import time

import numpy as np
import torch

from nemo.utils import logging
from nemo.utils.app_state import AppState


class BlendableDataset(torch.utils.data.Dataset):
    """ """

    def __init__(self, datasets, weights, size):
        self.datasets = datasets
        num_datasets = len(datasets)
        assert num_datasets == len(weights)

        self.size = size

        # Normalize weights.
        weights = np.array(weights, dtype=np.float64)
        sum_weights = np.sum(weights)
        assert sum_weights > 0.0
        weights /= sum_weights

        # Build indecies.
        start_time = time.time()
        assert num_datasets < 255
        self.dataset_index = np.zeros(self.size, dtype=np.uint8)
        self.dataset_sample_index = np.zeros(self.size, dtype=np.int64)

        app_state = AppState()
        try:
            if app_state.local_rank == 0:
                compile_helper()
            torch.distributed.barrier()
            from nemo.collections.common.data import helpers
        except ImportError:
            raise ImportError(
                'Could not compile megatron dataset C++ helper functions and therefore '
                'cannot import helpers python file.'
            )

        helpers.build_blending_indices(
            self.dataset_index,
            self.dataset_sample_index,
            weights,
            num_datasets,
            self.size,
            torch.distributed.get_rank() == 0,
        )
        logging.info(
            '> elapsed time for building blendable dataset indices: ' '{:.2f} (sec)'.format(time.time() - start_time)
        )

    def __len__(self):
        return self.size

    def __getitem__(self, idx):
        dataset_idx = self.dataset_index[idx]
        sample_idx = self.dataset_sample_index[idx]
        dataset_size = len(self.datasets[dataset_idx])
        # Ensure the sample index doesn't exceed the dataset size
        if sample_idx >= dataset_size:
            logging.warning(f"Index {sample_idx} out of bounds for dataset {dataset_idx}. Reusing existing examples.")
            sample_idx = sample_idx % dataset_size
            logging.warning(f"Reusing index {sample_idx} for dataset {dataset_idx}.")

        return self.datasets[dataset_idx][sample_idx]

    def create_data_mmap(self):
        """ """
        for dataset in self.datasets:
            dataset.create_data_mmap()


class MemoryEfficientBlendableDataset(torch.utils.data.Dataset):
    """
    A BlendableDataset implementation that uses less memory than the original implementation.
    Indices are computed algorithmically instead of storing them in memory.

    To test call: MemoryEfficientBlendableDataset.test_index_blending()
    """

    def __init__(self, datasets, weights, size, weight_bins=100):
        self.datasets = datasets
        num_datasets = len(datasets)
        assert num_datasets == len(weights)

        weight_bins = min(weight_bins, size)

        self.size = size
        self.weight_bins = weight_bins

        # Normalize weights.
        weights = np.array(weights, dtype=np.float64)
        assert (weights > 0.0).all()
        sum_weights = np.sum(weights)
        assert sum_weights > 0.0
        self.weights = weights / sum_weights

        # create ds index based on weights
        ds_index = []
        ds_bias = []
        for i, w in enumerate(self.weights):
            n = int(w * weight_bins)
            ds_index.extend([i] * n)
            ds_bias.extend(range(n))
        # make sure arrays have length of weight_bins
        n = weight_bins - len(ds_index)
        ds_index.extend([i] * n)
        ds_bias.extend(range(ds_bias[-1], ds_bias[-1] + n))

        self.ds_index = np.array(ds_index, dtype=np.uint32)
        self.ds_index_size = np.array([(self.ds_index == i).sum() for i in range(num_datasets)], dtype=np.uint32)
        assert (self.ds_index_size > 0).all(), (
            "Some datasets have no samples in the blendable dataset, "
            "increase weight_bins or the offending weight. "
            f"ds_index_size = {self.ds_index_size}"
        )
        self.ds_bias = np.array(ds_bias, dtype=np.uint32)

        self.ds_size = np.array([len(ds) for ds in datasets], dtype=np.uint32)

    def get_ds_sample_idx(self, idx):
        """Returns ds index and sample index (within the ds) for the given index in the blendable dataset."""

        bin = idx % self.weight_bins
        ds_idx = self.ds_index[bin]
        sample_idx = (self.ds_bias[bin] + (idx // self.weight_bins) * self.ds_index_size[ds_idx]) % self.ds_size[
            ds_idx
        ]

        return ds_idx, sample_idx

    def __len__(self):
        return self.size

    def __getitem__(self, idx):
        ds_idx, sample_idx = self.get_ds_sample_idx(idx)

        return self.datasets[ds_idx][sample_idx]

    @classmethod
    def test_index_blending(cls):
        """Visualize indices of blended dataset"""

        import matplotlib.pyplot as plt

        plt.ion()

        class DS(torch.utils.data.Dataset):
            """ """

            def __init__(self, size, data):
                self.size = size
                self.data = data

            def __len__(self):
                return self.size

            def __getitem__(self, idx):
                return self.data[idx]

        for weight_bins in [10, 100]:
            blend_ds = MemoryEfficientBlendableDataset(
                [DS(10, "a"), DS(10, "b"), DS(10, "c")], [0.5, 0.3, 0.2], 50, weight_bins=weight_bins
            )

            ds_sample_idx_list = [blend_ds.get_ds_sample_idx(i) for i in range(50)]
            ds_list = list(zip(*ds_sample_idx_list))[0]
            sample_list = list(zip(*ds_sample_idx_list))[1]

            plt.figure()
            plt.plot(ds_list, label="ds idx")
            plt.plot(sample_list, label="sample")
            plt.legend()
            plt.grid()
            plt.title(f"weight_bins={weight_bins}")


def compile_helper():
    """Compile helper function ar runtime. Make sure this
    is invoked on a single process."""

    path = os.path.abspath(os.path.dirname(__file__))
    ret = subprocess.run(['make', '-C', path])
    if ret.returncode != 0:
        logging.error("Making C++ dataset helpers module failed, exiting.")
        import sys

        sys.exit(1)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import lhotse
from torch.utils.data import DataLoader, IterableDataset
from nemo.lightning.pytorch.plugins import MegatronDataSampler
from nemo.utils import logging


class SpeechLMDataSampler(MegatronDataSampler):
    """
    Overwrite the default MegatronDataSampler to not add batch sampler when iteralbe dataset is used.
    """

    def transform_dataloader(self, dataloader: DataLoader, consumed_samples: int = 0) -> DataLoader:
        """
        Overwrites the MegatronDataSampler.transform_dataloader() function
        """
        sampler = getattr(dataloader, 'sampler', None)

        if isinstance(sampler, lhotse.dataset.sampling.base.CutSampler):
            logging.info(f"Using Lhotse sampler for dataloader {dataloader}. Skipping Megatron data sampler.")
            return dataloader

        try:
            _ = len(dataloader.dataset)
            has_len = True
        except TypeError:
            has_len = False

        if not has_len:
            logging.info(f"Dataset {dataloader.dataset} does not have __len__ method. Skipping Megatron data sampler.")
            return dataloader

        if isinstance(dataloader.dataset, IterableDataset):
            logging.info(f"Dataset {dataloader.dataset} is an IterableDataset. Skipping Megatron data sampler.")
            return dataloader

        return super().transform_dataloader(dataloader, consumed_samples)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
from typing import Any, Dict, List, Optional

import nemo_run as run
from nemo_run.config import get_nemorun_home
from nemo_run.core.execution.launcher import SlurmTemplate

from nemo.lightning.base import DEFAULT_NEMO_CACHE_HOME
from nemo.utils import logging

DEFAULT_NEMO_HOME = os.getenv('NEMO_HOME', DEFAULT_NEMO_CACHE_HOME)

# NOTE: If you update this template,
# PLEASE test it by submitting a job to GPU/node/cluster and verifying the sbatch and bash scripts.
INLINE_TEMPLATE = r"""
#!/usr/bin/env bash
set -euo pipefail

# NOTE: DO NOT change the single quotes to double quotes.
bash -c '{{ pre_cmds }} {{ command }}'
"""


def slurm_executor(
    gpu: str,
    account: str,
    partition: str,
    log_dir: str,
    nodes: int,
    num_gpus_per_node: int,
    time_limit: str = "00:30:00",
    container_image: str = "nvcr.io/nvidia/nemo:dev",
    custom_mounts: List[str] = [],
    custom_env_vars: Dict[str, str] = {},
    custom_srun_args: List[str] = [],
    hf_token: str = None,
    nemo_home: str = DEFAULT_NEMO_HOME,
    wandb_key: str = None,
    network: str = None,
    custom_bash_cmds: List[str] = None,
    optional_gpus_per_node: Optional[int] = None,
    additional_slurm_params: Dict[str, Any] = None,
) -> run.SlurmExecutor:
    """
    Slurm cluster definition with appropriate cluster params and NeMo container params needed for pre-training
    and fine-tuning experiments

    Args:
        additional_slurm_params: Dict[str, Any], optional
            Additional SLURM parameters to pass to sbatch. These will be converted to #SBATCH directives.
            Example: {"nodelist": "node001,node002", "constraint": "gpu"} will generate:
                #SBATCH --nodelist=node001,node002
                #SBATCH --constraint=gpu
    """
    PERF_ENV_VARS = {
        "TORCH_NCCL_AVOID_RECORD_STREAMS": "1",  # Disable caching NCCL communication buffer memory
        "HF_HUB_OFFLINE": "1",  # Enable online downloads from HuggingFace
        "TOKENIZERS_PARALLELISM": "False",  # Restrict warning message prints
        "NCCL_NVLS_ENABLE": "0",  # Disable NVLink SHARP to save memory
        "NVTE_FLASH_ATTN": "1",  # Enable Flash Attention, which is needed to enable cuDNN fused attention
        "NVTE_FUSED_ATTN": "1",  # Enable cuDNN fused attention
        "NEMO_LOG_MEMORY_USAGE": "1",  # Print memory allocation
    }

    custom_bash_cmds = [] if custom_bash_cmds is None else custom_bash_cmds
    err_msgs = []
    mounts = []
    srun_args = custom_srun_args.copy() + ["--mpi=pmix", "--no-container-mount-home", "--container-writable"]

    if log_dir != get_nemorun_home():
        err_msgs.append(f"\nRun `export NEMORUN_HOME={log_dir}` in your shell environment and rerun this script.")
    if len(err_msgs) > 0:
        logging.error("\n".join(err_msgs))
        sys.exit(1)

    if gpu.lower() not in ['b200']:
        # TODO: we currently disable PYTORCH_CUDA_ALLOC_CONF="expandable_segments:True"
        # on B200 as it causes an unexpected error. Add back when issue is debugged and fixed.
        PERF_ENV_VARS["PYTORCH_CUDA_ALLOC_CONF"] = "expandable_segments:True"
    PERF_ENV_VARS["NEMORUN_HOME"] = log_dir
    if wandb_key is not None:
        PERF_ENV_VARS["WANDB_API_KEY"] = wandb_key

    if gpu.lower() == 'gb200':
        PERF_ENV_VARS["NCCL_NET_GDR_LEVEL"] = "PHB"  # For NCCL 2.25
        PERF_ENV_VARS["NCCL_NET_GDR_C2C"] = "1"  # For NCCL 2.26

    if nemo_home != DEFAULT_NEMO_CACHE_HOME:  # DO NOT change this to 'DEFAULT_NEMO_HOME'/'NEMO_HOME'
        PERF_ENV_VARS["NEMO_HOME"] = nemo_home
        mounts.extend([f"{nemo_home}:{nemo_home}"])
    if hf_token is not None:
        PERF_ENV_VARS.update({"HF_TOKEN": hf_token, "HF_HUB_OFFLINE": "0"})

    PERF_ENV_VARS |= custom_env_vars
    mounts.extend(custom_mounts)

    # add --segment flag to sbatch if job uses GB200 and goes beyond one rack.
    segment = None
    if num_gpus_per_node == 4 and nodes > 18:
        for segment_candidate in range(18, 0, -1):
            if nodes % segment_candidate == 0:
                segment = segment_candidate
                break

    numa_divisor = 2 if gpu.lower() == 'gb200' else 4
    numa_cmd = f"numactl --cpunodebind=$((SLURM_LOCALID/{numa_divisor})) --membind=$((SLURM_LOCALID/{numa_divisor}))"
    custom_bash_cmds.append(numa_cmd)

    launcher = SlurmTemplate(
        template_inline=INLINE_TEMPLATE,
        template_vars={"pre_cmds": " ; ".join(custom_bash_cmds)},
    )

    executor = run.SlurmExecutor(
        account=account,
        partition=partition,
        tunnel=run.LocalTunnel(job_dir=os.path.join(log_dir, "experiments")),
        nodes=nodes,
        ntasks_per_node=num_gpus_per_node,
        gpus_per_node=optional_gpus_per_node,
        container_image=container_image,
        container_mounts=mounts,
        env_vars=PERF_ENV_VARS,
        srun_args=srun_args,
        time=time_limit,
        mem="0",
        exclusive=True,
        packager=run.GitArchivePackager(),
        segment=segment,
        network=network,
        launcher=launcher,
        additional_parameters=additional_slurm_params,
    )

    return executor



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from lightning.pytorch.callbacks.progress import TQDMProgressBar
from lightning.pytorch.callbacks.progress.tqdm_progress import _update_n


class MegatronProgressBar(TQDMProgressBar):
    """
    Add MegatronProgressBar to remove 's/it' and display progress per step instead of per microbatch
    for megatron models.
    """

    def init_train_tqdm(self):
        """
        Override bar_format to not have 's/it'.
        """
        self.bar = super().init_train_tqdm()
        self.bar.bar_format = "{desc}: {percentage:3.0f}%|{bar}| {n_fmt}/{total_fmt} [{elapsed}<{remaining}{postfix}]"
        return self.bar

    def on_train_epoch_start(self, trainer, *_):
        if trainer.max_steps > 0:  # and (trainer.ckpt_path is not None):
            # while resuming from a ckpt use trainer.max_steps as the total for progress bar as trainer.num_training_batches
            # is truncated to max_steps - step being resumed at
            num_training_batches = trainer.max_steps
        else:
            num_training_batches = trainer.num_training_batches

        self.train_progress_bar.reset(num_training_batches)
        self.train_progress_bar.initial = 0
        self.train_progress_bar.set_description(f"Epoch {trainer.current_epoch}")

    def on_train_batch_end(self, trainer, pl_module, *_, **__):
        """
        Override parent class on_train_batch_end to update progress bar per global batch instead of per microbatch.
        """
        n = trainer.strategy.current_epoch_step
        if self._should_update(n, self.train_progress_bar.total):
            _update_n(self.train_progress_bar, n)
            self.train_progress_bar.set_postfix(self.get_metrics(trainer, pl_module), refresh=False)


def calculate_data_parallel_groups() -> int:
    from nemo.utils import AppState

    app_state = AppState()

    pipeline_model_parallel_size = app_state.pipeline_model_parallel_size
    tensor_model_parallel_size = app_state.tensor_model_parallel_size

    world_size = app_state.world_size
    data_parallel_group_len = world_size // (pipeline_model_parallel_size * tensor_model_parallel_size)

    return world_size // data_parallel_group_len



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional

import lightning.pytorch as pl
import numpy as np
import torch
from lightning.pytorch.utilities.types import EVAL_DATALOADERS, TRAIN_DATALOADERS
from torch.utils import data
from torch.utils.data import DataLoader, Dataset

from nemo.lightning.pytorch.plugins import MegatronDataSampler
from nemo.utils import logging


class MockDataModule(pl.LightningDataModule):
    """
    Mock data module with data sampling and preprocessing configurations.
    """

    def __init__(
        self,
        seq_length: int = 77,
        decoder_seq_length: Optional[int] = None,
        tokenizer: Optional = None,
        image_processor: Optional = None,
        micro_batch_size: int = 4,
        global_batch_size: int = 8,
        rampup_batch_size: Optional[List[int]] = None,
        num_train_samples: int = 10_000_000,
        num_val_samples: int = 10_000_000,
        num_test_samples: int = 10_000_000,
        num_workers: int = 8,
        pin_memory: bool = True,
        persistent_workers: bool = False,
        task_encoder: Optional[Any] = None,
    ):
        """
        Initializes the mock data module with data sampling and preprocessing configurations.
        task_encoder: This Mock data module uses Energon Task encoder if provided.

        Args:
            seq_length (int): Maximum sequence length for tokens.
            decoder_seq_length (Optional[int]): Sequence length for the decoder. Used by Megatron Sampler.
            tokenizer: Tokenizer for text processing.
            image_processor: Processor for image preprocessing.
            micro_batch_size (int): Batch size for training and validation.
            global_batch_size (int): Total batch size across GPUs.
            rampup_batch_size (Optional[List[int]]): Batch size ramp-up schedule. Used by Megatron Sampler.
            num_train_samples (int): Number of training samples.
            num_val_samples (int): Number of validation samples.
            num_test_samples (int): Number of testing samples.
            num_workers (int): Number of workers for data loading.
            pin_memory (bool): Whether to pin memory for data loading.
            persistent_workers (bool): Whether workers should remain persistent.
            task_encoder: Task encoder for Energon tasks.
        """
        super().__init__()
        self.seq_length = seq_length
        self.decoder_seq_length = decoder_seq_length
        self.micro_batch_size = micro_batch_size
        self.global_batch_size = global_batch_size
        self.num_train_samples = num_train_samples
        self.num_val_samples = num_val_samples
        self.num_test_samples = num_test_samples
        self.num_workers = num_workers
        self.pin_memory = pin_memory
        self.persistent_workers = persistent_workers
        self.task_encoder = task_encoder
        self.tokenizer = tokenizer
        self.image_processor = image_processor

        if tokenizer is None or image_processor is None:
            logging.warning("Processor or tokenizer are not provided! Fall back to `openai/clip-vit-large-patch14`.")
            from transformers import AutoProcessor

            from nemo.collections.common.tokenizers.huggingface.auto_tokenizer import AutoTokenizer

            processor = AutoProcessor.from_pretrained("openai/clip-vit-large-patch14")
            self.tokenizer = tokenizer or AutoTokenizer("openai/clip-vit-large-patch14")
            self.image_processor = image_processor or processor.image_processor
        self.data_sampler = MegatronDataSampler(
            seq_len=self.seq_length,
            decoder_seq_len=self.decoder_seq_length,
            micro_batch_size=micro_batch_size,
            global_batch_size=global_batch_size,
            rampup_batch_size=rampup_batch_size,
        )

    def setup(self, stage: str = "") -> None:
        # pylint: disable=C0116
        self._train_ds = _MockClipDataset(
            self.tokenizer,
            self.image_processor,
            "train",
            self.num_train_samples,
            self.seq_length,
            task_encoder=self.task_encoder,
        )
        self._validation_ds = _MockClipDataset(
            self.tokenizer, self.image_processor, "valid", self.num_val_samples, self.seq_length
        )
        self._test_ds = _MockClipDataset(
            self.tokenizer, self.image_processor, "test", self.num_test_samples, self.seq_length
        )

    def train_dataloader(self) -> TRAIN_DATALOADERS:
        # pylint: disable=C0116
        if not hasattr(self, "_train_ds"):
            self.setup()
        return self._create_dataloader(self._train_ds)

    def val_dataloader(self) -> EVAL_DATALOADERS:
        # pylint: disable=C0116
        if not hasattr(self, "_validation_ds"):
            self.setup()
        return self._create_dataloader(self._validation_ds)

    def test_dataloader(self) -> EVAL_DATALOADERS:
        # pylint: disable=C0116
        if not hasattr(self, "_test_ds"):
            self.setup()
        return self._create_dataloader(self._test_ds)

    def _create_dataloader(self, dataset, **kwargs) -> DataLoader:
        # pylint: disable=C0116
        return DataLoader(
            dataset,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            persistent_workers=self.persistent_workers,
            batch_size=self.micro_batch_size,
            collate_fn=dataset.collate_fn,
            **kwargs,
        )

    def state_dict(self) -> Dict[str, Any]:
        """
        Save the state of the data module.

        This method is called when saving a checkpoint. It generates and saves the state of the data module,
        including the state of the dataloader and the number of consumed samples.

        Returns:
        Dict[str, Any]: A dictionary containing the state of the data module.
        """

        logging.warning("trainer object not connected to data module object returning empty state")
        return {}


class _MockClipDataset(Dataset):
    def __init__(
        self,
        tokenizer,
        image_processor,
        name: str,
        num_samples: int,
        seq_length: int,
        seed: int = 42,
        task_encoder=None,
    ) -> None:
        super().__init__()
        self.name = name
        self.seq_length = seq_length

        self.vocab_size = tokenizer.vocab_size

        crop_size = image_processor.crop_size
        self.image_height, self.image_width = crop_size["height"], crop_size["width"]

        self.length = num_samples
        self.seed = seed
        self.task_encoder = task_encoder

    def __len__(self) -> int:
        return self.length

    def __getitem__(self, idx) -> Dict[str, torch.Tensor]:
        # Generate data of the expected size and datatype (based on GPTDataset).

        np_gen = np.random.default_rng(seed=(self.seed + idx))
        tokens = torch.from_numpy(np_gen.integers(self.vocab_size, size=[self.seq_length], dtype=np.int64))
        images = torch.from_numpy(np_gen.random(size=[3, self.image_height, self.image_width], dtype=np.float32))

        if self.task_encoder is not None:
            # Use energon task encoder if provided
            return self.task_encoder.encode_sample({"image": images, "txt": "This is Random Mock Text"})

        return {
            "images": images,
            "captions": tokens,
        }

    def _collate_fn(self, batch):
        """
        A default implementation of a collation function.
        Users should override this method to define custom data loaders.
        """
        collated_batch = data.dataloader.default_collate(batch)
        collated_batch["attention_mask"] = None
        return collated_batch

    def collate_fn(self, batch):
        """Method that user pass as functor to DataLoader.

        The method optionally performs neural type checking and add types to the outputs.

        Please note, subclasses of Dataset should not implement `input_types`.

        # Usage:
        dataloader = torch.utils.data.DataLoader(
                ....,
                collate_fn=dataset.collate_fn,
                ....
        )

        Returns
        -------
            Collated batch, with or without types.
        """
        return self._collate_fn(batch)



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import numpy as np
import pytest
import torch

from nemo.collections.audio.losses.audio import (
    MAELoss,
    MSELoss,
    SDRLoss,
    calculate_mae_batch,
    calculate_mean,
    calculate_mse_batch,
    calculate_sdr_batch,
    convolution_invariant_target,
    scale_invariant_target,
)
from nemo.collections.audio.losses.maxine import CombinedLoss
from nemo.collections.audio.parts.utils.audio import (
    calculate_sdr_numpy,
    convolution_invariant_target_numpy,
    scale_invariant_target_numpy,
)


class TestAudioLosses:
    @pytest.mark.unit
    @pytest.mark.parametrize('num_channels', [1, 4])
    @pytest.mark.parametrize('use_mask', [True, False])
    @pytest.mark.parametrize('use_input_length', [True, False])
    def test_calculate_mean(self, num_channels: int, use_mask: bool, use_input_length: bool):
        """Test mean calculation"""
        batch_size = 8
        num_samples = 50
        num_batches = 10
        random_seed = 42
        eps = 1e-10
        atol = 1e-6

        rng = torch.Generator()
        rng.manual_seed(random_seed)

        for n in range(num_batches):

            input_signal = torch.randn(size=(batch_size, num_channels, num_samples), generator=rng)
            # Random input length
            input_length = torch.randint(low=1, high=num_samples, size=(batch_size,), generator=rng)
            # Corresponding mask
            mask = torch.zeros(size=(batch_size, 1, num_samples))
            for i in range(batch_size):
                mask[i, :, : input_length[i]] = 1.0

            if use_mask and use_input_length:
                with pytest.raises(RuntimeError):
                    calculate_mean(input_signal, mask=mask, input_length=input_length, eps=eps)
                # Done with this test
                continue

            if use_mask:
                uut = calculate_mean(input_signal, mask=mask, eps=eps)
            elif use_input_length:
                uut = calculate_mean(input_signal, input_length=input_length, eps=eps)
            else:
                uut = calculate_mean(input_signal, eps=eps)

            # Calculate mean manually
            for b in range(batch_size):
                for c in range(num_channels):
                    golden = torch.mean(
                        input_signal[b, c, : input_length[b] if use_input_length or use_mask else num_samples]
                    )
                    assert torch.allclose(
                        uut[b, c], golden, atol=atol
                    ), f"Mean not matching for example {n}, channel {c}"

    @pytest.mark.unit
    def test_calculate_sdr_scale_and_convolution_invariant(self):
        """Test SDR calculation with scale and conovolution invariant options."""
        estimate = torch.randn(size=(1, 1, 100))
        target = torch.randn(size=(1, 1, 100))

        with pytest.raises(ValueError):
            # using both scale and convolution invariant is not allowed
            calculate_sdr_batch(estimate=estimate, target=target, scale_invariant=True, convolution_invariant=True)

    @pytest.mark.unit
    def test_calculate_mse_input_and_mask(self):
        """Test MSE calculation with simultaneous input length and mask."""
        estimate = torch.randn(size=(1, 1, 100))
        target = torch.randn(size=(1, 1, 100))
        input_length = torch.tensor([100])
        mask = torch.ones(size=(1, 1, 100))

        with pytest.raises(RuntimeError):
            # using both input_length and mask is not allowed
            calculate_mse_batch(estimate=estimate, target=target, input_length=input_length, mask=mask)

    @pytest.mark.unit
    def test_calculate_mse_invalid_dimensions(self):
        """Test MSE calculation with unsupported dimensions."""
        estimate = torch.randn(size=(1, 1, 100, 10))
        target = torch.randn(size=(1, 1, 100))

        with pytest.raises(AssertionError):
            # mismatched dimensions are not allowed
            calculate_mse_batch(estimate=estimate, target=target)

        estimate = torch.randn(size=(1, 1, 100, 10, 20))
        target = torch.randn(size=estimate.shape)

        with pytest.raises(RuntimeError):
            # dimensions larger than four are not allowed
            calculate_mse_batch(estimate=estimate, target=target)

    @pytest.mark.unit
    def test_calculate_mae_input_and_mask(self):
        """Test MAE calculation with simultaneous input length and mask."""
        estimate = torch.randn(size=(1, 1, 100))
        target = torch.randn(size=(1, 1, 100))
        input_length = torch.tensor([100])
        mask = torch.ones(size=(1, 1, 100))

        with pytest.raises(RuntimeError):
            # using both input_length and mask is not allowed
            calculate_mae_batch(estimate=estimate, target=target, input_length=input_length, mask=mask)

    @pytest.mark.unit
    def test_calculate_mae_invalid_dimensions(self):
        """Test MAE calculation with unsupported dimensions."""
        estimate = torch.randn(size=(1, 1, 100, 10))
        target = torch.randn(size=(1, 1, 100))

        with pytest.raises(AssertionError):
            # mismatched dimensions are not allowed
            calculate_mae_batch(estimate=estimate, target=target)

        estimate = torch.randn(size=(1, 1, 100, 10, 20))
        target = torch.randn(size=estimate.shape)

        with pytest.raises(RuntimeError):
            # dimensions larger than four are not allowed
            calculate_mae_batch(estimate=estimate, target=target)

    @pytest.mark.unit
    @pytest.mark.parametrize('num_channels', [1, 4])
    def test_sdr(self, num_channels: int):
        """Test SDR calculation"""
        test_eps = [0, 1e-16, 1e-1]
        batch_size = 8
        num_samples = 50
        num_batches = 10
        random_seed = 42
        atol = 1e-6

        _rng = np.random.default_rng(seed=random_seed)

        for remove_mean in [True, False]:
            for eps in test_eps:

                sdr_loss = SDRLoss(eps=eps, remove_mean=remove_mean)

                for n in range(num_batches):

                    # Generate random signal
                    target = _rng.normal(size=(batch_size, num_channels, num_samples))
                    # Random noise + scaling
                    noise = _rng.uniform(low=0.01, high=1) * _rng.normal(size=(batch_size, num_channels, num_samples))
                    # Estimate
                    estimate = target + noise

                    # DC bias for both
                    target += _rng.uniform(low=-1, high=1)
                    estimate += _rng.uniform(low=-1, high=1)

                    # Tensors for testing the loss
                    tensor_estimate = torch.tensor(estimate)
                    tensor_target = torch.tensor(target)

                    # Reference SDR
                    golden_sdr = np.zeros((batch_size, num_channels))
                    for b in range(batch_size):
                        for m in range(num_channels):
                            golden_sdr[b, m] = calculate_sdr_numpy(
                                estimate=estimate[b, m, :],
                                target=target[b, m, :],
                                remove_mean=remove_mean,
                                eps=eps,
                            )

                    # Calculate SDR in torch
                    uut_sdr = calculate_sdr_batch(
                        estimate=tensor_estimate,
                        target=tensor_target,
                        remove_mean=remove_mean,
                        eps=eps,
                    )

                    # Calculate SDR loss
                    uut_sdr_loss = sdr_loss(estimate=tensor_estimate, target=tensor_target)

                    # Compare torch SDR vs numpy
                    assert np.allclose(
                        uut_sdr.cpu().detach().numpy(), golden_sdr, atol=atol
                    ), f'SDR not matching for example {n}, eps={eps}, remove_mean={remove_mean}'

                    # Compare SDR loss vs average of torch SDR
                    assert np.isclose(
                        uut_sdr_loss, -uut_sdr.mean(), atol=atol
                    ), f'SDRLoss not matching for example {n}, eps={eps}, remove_mean={remove_mean}'

    @pytest.mark.unit
    @pytest.mark.parametrize('num_channels', [1, 4])
    def test_sdr_weighted(self, num_channels: int):
        """Test SDR calculation with weighting for channels"""
        batch_size = 8
        num_samples = 50
        num_batches = 10
        random_seed = 42
        atol = 1e-6

        _rng = np.random.default_rng(seed=random_seed)

        channel_weight = _rng.uniform(low=0.01, high=1.0, size=num_channels)
        channel_weight = channel_weight / np.sum(channel_weight)
        sdr_loss = SDRLoss(weight=channel_weight)

        for n in range(num_batches):

            # Generate random signal
            target = _rng.normal(size=(batch_size, num_channels, num_samples))
            # Random noise + scaling
            noise = _rng.uniform(low=0.001, high=10) * _rng.normal(size=target.shape)
            # Estimate
            estimate = target + noise

            # Tensors for testing the loss
            tensor_estimate = torch.tensor(estimate)
            tensor_target = torch.tensor(target)

            # Reference SDR
            golden_sdr = 0
            for b in range(batch_size):
                sdr = [
                    calculate_sdr_numpy(estimate=estimate[b, m, :], target=target[b, m, :])
                    for m in range(num_channels)
                ]
                # weighted sum
                sdr = np.sum(np.array(sdr) * channel_weight)
                golden_sdr += sdr
            golden_sdr /= batch_size  # average over batch

            # Calculate SDR
            uut_sdr_loss = sdr_loss(estimate=tensor_estimate, target=tensor_target)

            # Compare
            assert np.allclose(
                uut_sdr_loss.cpu().detach().numpy(), -golden_sdr, atol=atol
            ), f'SDRLoss not matching for example {n}'

    @pytest.mark.unit
    @pytest.mark.parametrize('num_channels', [1, 4])
    def test_sdr_input_length(self, num_channels):
        """Test SDR calculation with input length."""
        batch_size = 8
        max_num_samples = 50
        num_batches = 10
        random_seed = 42
        atol = 1e-6

        _rng = np.random.default_rng(seed=random_seed)

        sdr_loss = SDRLoss()

        for n in range(num_batches):

            # Generate random signal
            target = _rng.normal(size=(batch_size, num_channels, max_num_samples))
            # Random noise + scaling
            noise = _rng.uniform(low=0.001, high=10) * _rng.normal(size=target.shape)
            # Estimate
            estimate = target + noise

            # Limit calculation to random input_length samples
            input_length = _rng.integers(low=1, high=max_num_samples, size=batch_size)

            # Tensors for testing the loss
            tensor_estimate = torch.tensor(estimate)
            tensor_target = torch.tensor(target)
            tensor_input_length = torch.tensor(input_length)

            # Reference SDR
            golden_sdr = 0
            for b, b_len in enumerate(input_length):
                sdr = [
                    calculate_sdr_numpy(estimate=estimate[b, m, :b_len], target=target[b, m, :b_len])
                    for m in range(num_channels)
                ]
                sdr = np.mean(np.array(sdr))
                golden_sdr += sdr
            golden_sdr /= batch_size  # average over batch

            # Calculate SDR
            uut_sdr_loss = sdr_loss(estimate=tensor_estimate, target=tensor_target, input_length=tensor_input_length)

            # Compare
            assert np.allclose(
                uut_sdr_loss.cpu().detach().numpy(), -golden_sdr, atol=atol
            ), f'SDRLoss not matching for example {n}'

    @pytest.mark.unit
    @pytest.mark.parametrize('num_channels', [1, 4])
    def test_sdr_scale_invariant(self, num_channels: int):
        """Test SDR calculation with scale invariant option."""
        batch_size = 8
        max_num_samples = 50
        num_batches = 10
        random_seed = 42
        atol = 1e-6

        _rng = np.random.default_rng(seed=random_seed)

        sdr_loss = SDRLoss(scale_invariant=True)

        for n in range(num_batches):

            # Generate random signal
            target = _rng.normal(size=(batch_size, num_channels, max_num_samples))
            # Random noise + scaling
            noise = _rng.uniform(low=0.001, high=10) * _rng.normal(size=target.shape)
            # Estimate
            estimate = target + noise

            # Limit calculation to random input_length samples
            input_length = _rng.integers(low=1, high=max_num_samples, size=batch_size)

            # Tensors for testing the loss
            tensor_estimate = torch.tensor(estimate)
            tensor_target = torch.tensor(target)
            tensor_input_length = torch.tensor(input_length)

            # Reference SDR
            golden_sdr = 0
            for b, b_len in enumerate(input_length):
                sdr = [
                    calculate_sdr_numpy(
                        estimate=estimate[b, m, :b_len], target=target[b, m, :b_len], scale_invariant=True
                    )
                    for m in range(num_channels)
                ]
                sdr = np.mean(np.array(sdr))
                golden_sdr += sdr
            golden_sdr /= batch_size  # average over batch

            # Calculate SDR loss
            uut_sdr_loss = sdr_loss(estimate=tensor_estimate, target=tensor_target, input_length=tensor_input_length)

            # Compare
            assert np.allclose(
                uut_sdr_loss.cpu().detach().numpy(), -golden_sdr, atol=atol
            ), f'SDRLoss not matching for example {n}'

    @pytest.mark.unit
    @pytest.mark.parametrize('num_channels', [1, 4])
    def test_sdr_binary_mask(self, num_channels):
        """Test SDR calculation with temporal mask."""
        batch_size = 8
        max_num_samples = 50
        num_batches = 10
        random_seed = 42
        atol = 1e-6

        _rng = np.random.default_rng(seed=random_seed)

        sdr_loss = SDRLoss()

        for n in range(num_batches):

            # Generate random signal
            target = _rng.normal(size=(batch_size, num_channels, max_num_samples))
            # Random noise + scaling
            noise = _rng.uniform(low=0.001, high=10) * _rng.normal(size=target.shape)
            # Estimate
            estimate = target + noise

            # Limit calculation to masked samples
            mask = _rng.integers(low=0, high=2, size=(batch_size, num_channels, max_num_samples))

            # Tensors for testing the loss
            tensor_estimate = torch.tensor(estimate)
            tensor_target = torch.tensor(target)
            tensor_mask = torch.tensor(mask)

            # Reference SDR
            golden_sdr = 0
            for b in range(batch_size):
                sdr = [
                    calculate_sdr_numpy(
                        estimate=estimate[b, m, mask[b, m, :] > 0], target=target[b, m, mask[b, m, :] > 0]
                    )
                    for m in range(num_channels)
                ]
                sdr = np.mean(np.array(sdr))
                golden_sdr += sdr
            golden_sdr /= batch_size  # average over batch

            # Calculate SDR loss
            uut_sdr_loss = sdr_loss(estimate=tensor_estimate, target=tensor_target, mask=tensor_mask)

            # Compare
            assert np.allclose(
                uut_sdr_loss.cpu().detach().numpy(), -golden_sdr, atol=atol
            ), f'SDRLoss not matching for example {n}'

    @pytest.mark.unit
    @pytest.mark.parametrize('num_channels', [1])
    @pytest.mark.parametrize('sdr_max', [10, 0])
    def test_sdr_max(self, num_channels: int, sdr_max: float):
        """Test SDR calculation with soft max threshold."""
        batch_size = 8
        max_num_samples = 50
        num_batches = 10
        random_seed = 42
        atol = 1e-6

        _rng = np.random.default_rng(seed=random_seed)

        sdr_loss = SDRLoss(sdr_max=sdr_max)

        for n in range(num_batches):

            # Generate random signal
            target = _rng.normal(size=(batch_size, num_channels, max_num_samples))
            # Random noise + scaling
            noise = _rng.uniform(low=0.001, high=10) * _rng.normal(size=target.shape)
            # Estimate
            estimate = target + noise

            # Limit calculation to random input_length samples
            input_length = _rng.integers(low=1, high=max_num_samples, size=batch_size)

            # Tensors for testing the loss
            tensor_estimate = torch.tensor(estimate)
            tensor_target = torch.tensor(target)
            tensor_input_length = torch.tensor(input_length)

            # Reference SDR
            golden_sdr = 0
            for b, b_len in enumerate(input_length):
                sdr = [
                    calculate_sdr_numpy(estimate=estimate[b, m, :b_len], target=target[b, m, :b_len], sdr_max=sdr_max)
                    for m in range(num_channels)
                ]
                sdr = np.mean(np.array(sdr))
                golden_sdr += sdr
            golden_sdr /= batch_size  # average over batch

            # Calculate SDR loss
            uut_sdr_loss = sdr_loss(estimate=tensor_estimate, target=tensor_target, input_length=tensor_input_length)

            # Compare
            assert np.allclose(
                uut_sdr_loss.cpu().detach().numpy(), -golden_sdr, atol=atol
            ), f'SDRLoss not matching for example {n}'

    @pytest.mark.unit
    @pytest.mark.parametrize('filter_length', [1, 32])
    @pytest.mark.parametrize('num_channels', [1, 4])
    @pytest.mark.parametrize('use_mask', [True, False])
    @pytest.mark.parametrize('use_input_length', [True, False])
    def test_target_calculation(self, num_channels: int, filter_length: int, use_mask: bool, use_input_length: bool):
        """Test target calculation with scale and convolution invariance."""
        batch_size = 8
        max_num_samples = 50
        num_batches = 10
        random_seed = 42
        atol = 1e-6

        _rng = np.random.default_rng(seed=random_seed)

        for n in range(num_batches):

            # Generate random signal
            target = _rng.normal(size=(batch_size, num_channels, max_num_samples))
            # Random noise + scaling
            noise = _rng.uniform(low=0.001, high=10) * _rng.normal(size=target.shape)
            # Estimate
            estimate = target + noise

            # Limit calculation to random input_length samples
            input_length = _rng.integers(low=filter_length, high=max_num_samples, size=batch_size)

            # Corresponding mask
            mask = torch.zeros(size=(batch_size, 1, max_num_samples))
            for i in range(batch_size):
                mask[i, :, : input_length[i]] = 1.0

            if use_mask and use_input_length:
                with pytest.raises(RuntimeError):
                    scale_invariant_target(
                        estimate=torch.tensor(estimate),
                        target=torch.tensor(target),
                        input_length=torch.tensor(input_length),
                        mask=mask,
                    )

                with pytest.raises(RuntimeError):
                    convolution_invariant_target(
                        estimate=torch.tensor(estimate),
                        target=torch.tensor(target),
                        input_length=torch.tensor(input_length),
                        mask=mask,
                        filter_length=filter_length,
                    )

                # Done with this test
                continue

            # UUT
            si_target = scale_invariant_target(
                estimate=torch.tensor(estimate),
                target=torch.tensor(target),
                input_length=torch.tensor(input_length) if use_input_length else None,
                mask=mask if use_mask else None,
            )
            ci_target = convolution_invariant_target(
                estimate=torch.tensor(estimate),
                target=torch.tensor(target),
                input_length=torch.tensor(input_length) if use_input_length else None,
                mask=mask if use_mask else None,
                filter_length=filter_length,
            )

            if filter_length == 1:
                assert torch.allclose(ci_target, si_target), f'SI and CI should match for filter_length=1'

            # Compare against numpy
            for b in range(batch_size):
                # valid length for the current example
                b_len = input_length[b] if use_input_length or use_mask else max_num_samples

                # calculate reference target for each channel
                for m in range(num_channels):
                    # Scale invariant reference
                    si_target_ref = scale_invariant_target_numpy(
                        estimate=estimate[b, m, :b_len], target=target[b, m, :b_len]
                    )

                    assert np.allclose(
                        si_target[b, m, :b_len].cpu().detach().numpy(), si_target_ref, atol=atol
                    ), f'SI not matching for example {n}, channel {m}'

                    # Convolution invariant reference
                    ci_target_ref = convolution_invariant_target_numpy(
                        estimate=estimate[b, m, :b_len], target=target[b, m, :b_len], filter_length=filter_length
                    )

                    assert np.allclose(
                        ci_target[b, m, :b_len].cpu().detach().numpy(), ci_target_ref, atol=atol
                    ), f'CI not matching for example {n}, channel {m}'

    @pytest.mark.unit
    @pytest.mark.parametrize('filter_length', [1, 32])
    @pytest.mark.parametrize('num_channels', [1, 4])
    def test_sdr_convolution_invariant(self, num_channels: int, filter_length: int):
        """Test SDR calculation with convolution invariant option."""
        batch_size = 8
        max_num_samples = 50
        num_batches = 10
        random_seed = 42
        atol = 1e-6

        _rng = np.random.default_rng(seed=random_seed)

        sdr_loss = SDRLoss(convolution_invariant=True, convolution_filter_length=filter_length)

        for n in range(num_batches):

            # Generate random signal
            target = _rng.normal(size=(batch_size, num_channels, max_num_samples))
            # Random noise + scaling
            noise = _rng.uniform(low=0.001, high=10) * _rng.normal(size=target.shape)
            # Estimate
            estimate = target + noise

            # Limit calculation to random input_length samples
            input_length = _rng.integers(low=filter_length, high=max_num_samples, size=batch_size)

            # Calculate SDR loss
            uut_sdr_loss = sdr_loss(
                estimate=torch.tensor(estimate), target=torch.tensor(target), input_length=torch.tensor(input_length)
            )

            # Reference SDR
            golden_sdr = 0
            for b, b_len in enumerate(input_length):
                sdr = [
                    calculate_sdr_numpy(
                        estimate=estimate[b, m, :b_len],
                        target=target[b, m, :b_len],
                        convolution_invariant=True,
                        convolution_filter_length=filter_length,
                    )
                    for m in range(num_channels)
                ]
                sdr = np.mean(np.array(sdr))
                golden_sdr += sdr
            golden_sdr /= batch_size  # average over batch

            # Compare
            assert np.allclose(
                uut_sdr_loss.cpu().detach().numpy(), -golden_sdr, atol=atol
            ), f'SDRLoss not matching for example {n}'

    @pytest.mark.unit
    def test_sdr_scale_and_convolution_invariant(self):
        """Test SDR calculation with scale and conovolution invariant options."""
        with pytest.raises(ValueError):
            # using both scale and convolution invariant is not allowed
            SDRLoss(scale_invariant=True, convolution_invariant=True)

    @pytest.mark.unit
    def test_sdr_length_and_mask(self):
        """Test SDR calculation with simultaneous input length and mask."""

        estimate = torch.randn(size=(1, 1, 100))
        target = torch.randn(size=(1, 1, 100))
        input_length = torch.tensor([100])
        mask = torch.ones(size=(1, 1, 100))

        sdr_loss = SDRLoss(scale_invariant=False, convolution_invariant=False)

        with pytest.raises(RuntimeError):
            # using both input_length and mask is not allowed
            sdr_loss(estimate=estimate, target=target, input_length=input_length, mask=mask)

    @pytest.mark.unit
    def test_sdr_invalid_weight(self):
        """Test SDR with invalid weights."""
        with pytest.raises(ValueError):
            # negative weights are not allowed
            SDRLoss(weight=[-1, 1])

        with pytest.raises(ValueError):
            # weights should sum to 1
            SDRLoss(weight=[0.1, 0.1])

    @pytest.mark.unit
    def test_sdr_invalid_reduction(self):
        """Test SDR with invalid reduction."""
        with pytest.raises(ValueError):
            SDRLoss(reduction='not-mean')

    @pytest.mark.unit
    @pytest.mark.parametrize('num_channels', [1, 4])
    @pytest.mark.parametrize('ndim', [3, 4])
    def test_mse(self, num_channels: int, ndim: int):
        """Test MSE calculation"""
        batch_size = 8
        num_samples = 50
        num_features = 123
        num_batches = 10
        random_seed = 42
        atol = 1e-6

        signal_shape = (
            (batch_size, num_channels, num_features, num_samples)
            if ndim == 4
            else (batch_size, num_channels, num_samples)
        )

        reduction_dim = (-2, -1) if ndim == 4 else -1

        mse_loss = MSELoss(ndim=ndim)

        _rng = np.random.default_rng(seed=random_seed)

        for n in range(num_batches):

            # Generate random signal
            target = _rng.normal(size=signal_shape)
            # Random noise + scaling
            noise = _rng.uniform(low=0.01, high=1) * _rng.normal(size=signal_shape)
            # Estimate
            estimate = target + noise

            # DC bias for both
            target += _rng.uniform(low=-1, high=1)
            estimate += _rng.uniform(low=-1, high=1)

            # Tensors for testing the loss
            tensor_estimate = torch.tensor(estimate)
            tensor_target = torch.tensor(target)

            # Reference MSE
            golden_mse = np.zeros((batch_size, num_channels))
            for b in range(batch_size):
                for m in range(num_channels):
                    err = estimate[b, m, :] - target[b, m, :]
                    golden_mse[b, m] = np.mean(np.abs(err) ** 2, axis=reduction_dim)

            # Calculate MSE in torch
            uut_mse = calculate_mse_batch(estimate=tensor_estimate, target=tensor_target)

            # Calculate MSE loss
            uut_mse_loss = mse_loss(estimate=tensor_estimate, target=tensor_target)

            # Compare torch SDR vs numpy
            assert np.allclose(
                uut_mse.cpu().detach().numpy(), golden_mse, atol=atol
            ), f'MSE not matching for example {n}'

            # Compare SDR loss vs average of torch SDR
            assert np.isclose(uut_mse_loss, uut_mse.mean(), atol=atol), f'MSELoss not matching for example {n}'

    @pytest.mark.unit
    @pytest.mark.parametrize('num_channels', [1, 4])
    @pytest.mark.parametrize('ndim', [3, 4])
    def test_mse_weighted(self, num_channels: int, ndim: int):
        """Test MSE calculation with weighting for channels"""
        batch_size = 8
        num_samples = 50
        num_features = 123
        num_batches = 10
        random_seed = 42
        atol = 1e-6

        signal_shape = (
            (batch_size, num_channels, num_features, num_samples)
            if ndim == 4
            else (batch_size, num_channels, num_samples)
        )

        reduction_dim = (-2, -1) if ndim == 4 else -1

        _rng = np.random.default_rng(seed=random_seed)

        channel_weight = _rng.uniform(low=0.01, high=1.0, size=num_channels)
        channel_weight = channel_weight / np.sum(channel_weight)
        mse_loss = MSELoss(weight=channel_weight, ndim=ndim)

        for n in range(num_batches):

            # Generate random signal
            target = _rng.normal(size=signal_shape)
            # Random noise + scaling
            noise = _rng.uniform(low=0.001, high=10) * _rng.normal(size=target.shape)
            # Estimate
            estimate = target + noise

            # Tensors for testing the loss
            tensor_estimate = torch.tensor(estimate)
            tensor_target = torch.tensor(target)

            # Reference MSE
            golden_mse = 0
            for b in range(batch_size):
                mse = [
                    np.mean(np.abs(estimate[b, m, :] - target[b, m, :]) ** 2, axis=reduction_dim)
                    for m in range(num_channels)
                ]
                # weighted sum
                mse = np.sum(np.array(mse) * channel_weight)
                golden_mse += mse
            golden_mse /= batch_size  # average over batch

            # Calculate MSE loss
            uut_mse_loss = mse_loss(estimate=tensor_estimate, target=tensor_target)

            # Compare
            assert np.allclose(
                uut_mse_loss.cpu().detach().numpy(), golden_mse, atol=atol
            ), f'MSELoss not matching for example {n}'

    @pytest.mark.unit
    @pytest.mark.parametrize('num_channels', [1, 4])
    @pytest.mark.parametrize('ndim', [3, 4])
    def test_mse_input_length(self, num_channels: int, ndim: int):
        """Test MSE calculation with input length."""
        batch_size = 8
        max_num_samples = 50
        num_features = 123
        num_batches = 10
        random_seed = 42
        atol = 1e-6

        signal_shape = (
            (batch_size, num_channels, num_features, max_num_samples)
            if ndim == 4
            else (batch_size, num_channels, max_num_samples)
        )

        reduction_dim = (-2, -1) if ndim == 4 else -1

        _rng = np.random.default_rng(seed=random_seed)

        mse_loss = MSELoss(ndim=ndim)

        for n in range(num_batches):

            # Generate random signal
            target = _rng.normal(size=signal_shape)
            # Random noise + scaling
            noise = _rng.uniform(low=0.001, high=10) * _rng.normal(size=target.shape)
            # Estimate
            estimate = target + noise

            # Limit calculation to random input_length samples
            input_length = _rng.integers(low=1, high=max_num_samples, size=batch_size)

            # Tensors for testing the loss
            tensor_estimate = torch.tensor(estimate)
            tensor_target = torch.tensor(target)
            tensor_input_length = torch.tensor(input_length)

            # Reference MSE
            golden_mse = 0
            for b, b_len in enumerate(input_length):
                mse = [
                    np.mean(np.abs(estimate[b, m, ..., :b_len] - target[b, m, ..., :b_len]) ** 2, axis=reduction_dim)
                    for m in range(num_channels)
                ]
                mse = np.mean(np.array(mse))
                golden_mse += mse
            golden_mse /= batch_size  # average over batch

            # Calculate MSE
            uut_mse_loss = mse_loss(estimate=tensor_estimate, target=tensor_target, input_length=tensor_input_length)

            # Compare
            assert np.allclose(
                uut_mse_loss.cpu().detach().numpy(), golden_mse, atol=atol
            ), f'MSELoss not matching for example {n}'

    @pytest.mark.unit
    def test_mse_invalid_weight(self):
        """Test MSE with unsupported weights."""
        with pytest.raises(ValueError):
            # negative weights are not allowed
            MSELoss(weight=[-1, 1])

        with pytest.raises(ValueError):
            # weights should sum to 1
            MSELoss(weight=[0.1, 0.1])

    @pytest.mark.unit
    def test_mse_invalid_reduction(self):
        """Test MSE with unsupported reduction."""
        with pytest.raises(ValueError):
            # unsupported reduction
            MSELoss(reduction='not-mean')

    @pytest.mark.unit
    def test_mse_invalid_ndim(self):
        """Test MSE with unsupported dimensions."""
        with pytest.raises(ValueError):
            # supports dimensions 3, 4
            MSELoss(ndim=2)

        with pytest.raises(ValueError):
            # supports dimensions 3, 4
            MSELoss(ndim=5)

    @pytest.mark.unit
    @pytest.mark.parametrize('num_channels', [1, 4])
    @pytest.mark.parametrize('ndim', [3, 4])
    def test_mae(self, num_channels: int, ndim: int):
        """Test MAE calculation"""
        batch_size = 8
        num_samples = 50
        num_features = 123
        num_batches = 10
        random_seed = 42
        atol = 1e-6

        signal_shape = (
            (batch_size, num_channels, num_features, num_samples)
            if ndim == 4
            else (batch_size, num_channels, num_samples)
        )

        reduction_dim = (-2, -1) if ndim == 4 else -1

        mae_loss = MAELoss(ndim=ndim)

        _rng = np.random.default_rng(seed=random_seed)

        for n in range(num_batches):

            # Generate random signal
            target = _rng.normal(size=signal_shape)
            # Random noise + scaling
            noise = _rng.uniform(low=0.01, high=1) * _rng.normal(size=signal_shape)
            # Estimate
            estimate = target + noise

            # DC bias for both
            target += _rng.uniform(low=-1, high=1)
            estimate += _rng.uniform(low=-1, high=1)

            # Tensors for testing the loss
            tensor_estimate = torch.tensor(estimate)
            tensor_target = torch.tensor(target)

            # Reference MSE
            golden_mae = np.zeros((batch_size, num_channels))
            for b in range(batch_size):
                for m in range(num_channels):
                    err = estimate[b, m, :] - target[b, m, :]
                    golden_mae[b, m] = np.mean(np.abs(err), axis=reduction_dim)

            # Calculate MSE loss
            uut_mae_loss = mae_loss(estimate=tensor_estimate, target=tensor_target)

            # Compare
            assert np.allclose(
                uut_mae_loss.cpu().detach().numpy(), golden_mae.mean(), atol=atol
            ), f'MAE not matching for example {n}'

    @pytest.mark.unit
    @pytest.mark.parametrize('num_channels', [1, 4])
    @pytest.mark.parametrize('ndim', [3, 4])
    def test_mae_weighted(self, num_channels: int, ndim: int):
        """Test MAE calculation with weighting for channels"""
        batch_size = 8
        num_samples = 50
        num_features = 123
        num_batches = 10
        random_seed = 42
        atol = 1e-6

        signal_shape = (
            (batch_size, num_channels, num_features, num_samples)
            if ndim == 4
            else (batch_size, num_channels, num_samples)
        )

        reduction_dim = (-2, -1) if ndim == 4 else -1

        _rng = np.random.default_rng(seed=random_seed)

        channel_weight = _rng.uniform(low=0.01, high=1.0, size=num_channels)
        channel_weight = channel_weight / np.sum(channel_weight)
        mae_loss = MAELoss(weight=channel_weight, ndim=ndim)

        for n in range(num_batches):

            # Generate random signal
            target = _rng.normal(size=signal_shape)
            # Random noise + scaling
            noise = _rng.uniform(low=0.001, high=10) * _rng.normal(size=target.shape)
            # Estimate
            estimate = target + noise

            # Tensors for testing the loss
            tensor_estimate = torch.tensor(estimate)
            tensor_target = torch.tensor(target)

            # Reference MAE
            golden_mae = 0
            for b in range(batch_size):
                mae = [
                    np.mean(np.abs(estimate[b, m, :] - target[b, m, :]), axis=reduction_dim)
                    for m in range(num_channels)
                ]
                # weighted sum
                mae = np.sum(np.array(mae) * channel_weight)
                golden_mae += mae
            golden_mae /= batch_size  # average over batch

            # Calculate MAE loss
            uut_mae_loss = mae_loss(estimate=tensor_estimate, target=tensor_target)

            # Compare
            assert np.allclose(
                uut_mae_loss.cpu().detach().numpy(), golden_mae, atol=atol
            ), f'MAELoss not matching for example {n}'

    @pytest.mark.unit
    @pytest.mark.parametrize('num_channels', [1, 4])
    @pytest.mark.parametrize('ndim', [3, 4])
    def test_mae_input_length(self, num_channels: int, ndim: int):
        """Test MAE calculation with input length."""
        batch_size = 8
        max_num_samples = 50
        num_features = 123
        num_batches = 10
        random_seed = 42
        atol = 1e-6

        signal_shape = (
            (batch_size, num_channels, num_features, max_num_samples)
            if ndim == 4
            else (batch_size, num_channels, max_num_samples)
        )

        reduction_dim = (-2, -1) if ndim == 4 else -1

        _rng = np.random.default_rng(seed=random_seed)

        mae_loss = MAELoss(ndim=ndim)

        for n in range(num_batches):

            # Generate random signal
            target = _rng.normal(size=signal_shape)
            # Random noise + scaling
            noise = _rng.uniform(low=0.001, high=10) * _rng.normal(size=target.shape)
            # Estimate
            estimate = target + noise

            # Limit calculation to random input_length samples
            input_length = _rng.integers(low=1, high=max_num_samples, size=batch_size)

            # Tensors for testing the loss
            tensor_estimate = torch.tensor(estimate)
            tensor_target = torch.tensor(target)
            tensor_input_length = torch.tensor(input_length)

            # Reference MSE
            golden_mae = 0
            for b, b_len in enumerate(input_length):
                mae = [
                    np.mean(np.abs(estimate[b, m, ..., :b_len] - target[b, m, ..., :b_len]), axis=reduction_dim)
                    for m in range(num_channels)
                ]
                mae = np.mean(np.array(mae))
                golden_mae += mae
            golden_mae /= batch_size  # average over batch

            # Calculate MSE
            uut_mae_loss = mae_loss(estimate=tensor_estimate, target=tensor_target, input_length=tensor_input_length)

            # Compare
            assert np.allclose(
                uut_mae_loss.cpu().detach().numpy(), golden_mae, atol=atol
            ), f'MAELoss not matching for example {n}'

    @pytest.mark.unit
    def test_mae_invalid_weight(self):
        """Test MAE with invalid weights."""
        with pytest.raises(ValueError):
            # negative weights are not allowed
            MAELoss(weight=[-1, 1])

        with pytest.raises(ValueError):
            # weights should sum to 1
            MAELoss(weight=[0.1, 0.1])

    @pytest.mark.unit
    def test_mae_invalid_reduction(self):
        """Test MAE with invalid reduction."""
        with pytest.raises(ValueError):
            # unsupported reduction
            MAELoss(reduction='not-mean')

    @pytest.mark.unit
    def test_mae_invalid_ndim(self):
        """Test MAE with invalid dimensions."""
        with pytest.raises(ValueError):
            # supports dimensions 3, 4
            MAELoss(ndim=2)

        with pytest.raises(ValueError):
            # supports dimensions 3, 4
            MAELoss(ndim=5)

    @pytest.mark.unit
    def test_maxine_combined_loss(self, test_data_dir):
        INPUT_LOCATION = os.path.join(test_data_dir, 'audio', 'maxine', 'input.bin')
        ATOL = 1e-2

        GOLDEN_VALUES = [
            ((1, 0, 0, True, True), 80.0),
            ((0, 1, 0, True, True), 1.9749),
            ((0, 0, 1, True, True), 19.2192),
            ((1, 1, 1, True, True), 101.1941),
        ]
        batch_size = 16

        for value in GOLDEN_VALUES:
            config, golden = value
            sisnr_wt, asr_wt, spec_wt, use_asr, use_spec = config

            device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

            loss_instance = CombinedLoss(
                sample_rate=16000,
                fft_length=1920,
                hop_length=480,
                num_mels=320,
                sisnr_loss_weight=sisnr_wt,
                asr_loss_weight=asr_wt,
                spectral_loss_weight=spec_wt,
                use_asr_loss=use_asr,
                use_mel_spec=use_spec,
            ).to(device)

            input_data = torch.tensor(np.fromfile(INPUT_LOCATION, np.float32))
            input_data = input_data.repeat(batch_size).reshape((batch_size, 1, -1))
            input_data = input_data.to(device)

            estimate = torch.tensor(np.zeros(input_data.shape, np.float32)).reshape((batch_size, 1, -1)).to(device)

            loss = loss_instance.forward(estimate=estimate, target=input_data).cpu()

            assert np.allclose(loss, golden, atol=ATOL)
